from typing import Optional, Dict, Callable

import numpy as np

from ..util.parallel import handle_n_jobs


class SystemBase:

    def __init__(self, system, h: float, n_steps: int, props: Optional[Dict] = None):
        r""" Creates a new wrapper. """
        self._impl = system
        self._impl.h = h
        self._impl.n_steps = n_steps
        if props is not None:
            for prop in props:
                setattr(self._impl, prop, props[prop])

    @property
    def h(self) -> float:
        r""" The integration step size. """
        return self._impl.h

    @property
    def n_steps(self) -> int:
        r""" The number of integration steps between each evaluation of the state. """
        return self._impl.n_steps

    @property
    def dimension(self) -> int:
        r""" The dimension of the system's state.

        :type: int
        """
        return self._impl.dimension

    @property
    def time_dependent(self) -> bool:
        r""" Whether the potential (if available) and the right-hand side depend on time.

        :type: bool
        """
        return self._impl.time_dependent

    @property
    def integrator(self) -> str:
        r""" The type of integrator.

        :type: str
        """
        return self._impl.integrator

    @property
    def has_potential_function(self) -> bool:
        r""" Whether the system defines a potential.
        This means that the deterministic part of the right-hand side is the negative gradient of that potential.
        """
        return self._impl.has_potential_function

    @property
    def f(self) -> Callable[[float, np.ndarray], np.ndarray]:
        r""" The right-hand side of the system as callable function. In case of SDEs, this only evaluates the
        deterministic part of the function definition.

        Some of these definitions are vectorized.

        Returns
        -------
        right_hand_side : Callable[[float, ndarray], ndarray]
            The right-hand side.
        """
        return lambda t, x: self._impl.rhs(t, x).squeeze()

    @property
    def vectorized_f(self) -> bool:
        r""" Yields whether the right-hand side function is vectorized.

        :type: bool
        """
        return self._impl.vectorized_rhs


class _TimeIndependentRhsMixin:
    def trajectory(self: SystemBase, x0, length, seed=-1, n_jobs=None):
        r""" Simulates one or multiple trajectories.

        Parameters
        ----------
        x0 : array_like
            The initial condition. Must be compatible in shape to a (n_test_points, dimension)-array.
        length : int
            The length of the trajectory that is to be generated.
        seed : int, optional, default=-1
            The random seed. In case it is specified to be something else than `-1`, n_jobs must be set to `n_jobs=1`.
        n_jobs : int, optional, default=None
            Specify number of jobs according to :meth:`deeptime.util.parallel.handle_n_jobs`.

        Returns
        -------
        trajectory : np.ndarray
            The trajectory of shape (n_initial_conditions, n_evaluations, dimension).
            In case of just one initial condition the shape is squeezed to (n_evaluations, dimension).
        """
        n_jobs = handle_n_jobs(n_jobs)
        x0 = np.array(x0).reshape((-1, self._impl.dimension))
        n_initial_conditions = x0.shape[0]
        _, traj = self._impl.trajectory(x0, length, seed, n_jobs)
        if n_initial_conditions == 1:
            traj = traj[0]
        return traj

    def __call__(self: SystemBase, test_points, seed=-1, n_jobs=None):
        r"""Evolves the provided tests points under the dynamic for n_steps and returns.

        Parameters
        ----------
        test_points : array_like
            The test points.
        seed : int, optional, default=-1
            The seed for reproducibility. In case it is set to `seed >= 0` the number of jobs needs
            to be fixed to `n_jobs=1`.
        n_jobs : int, optional, default=None
            Specify number of jobs according to :meth:`deeptime.util.parallel.handle_n_jobs`.

        Returns
        -------
        points : np.ndarray
            The evolved test points.
        """
        n_jobs = handle_n_jobs(n_jobs)
        test_points = np.array(test_points).reshape((-1, self.dimension))
        return self._impl(test_points, seed, n_jobs)

    def potential(self: SystemBase, points):
        r""" Evaluates the system's potential function at given points in state space.

        Parameters
        ----------
        points : array_like
            The points.

        Returns
        -------
        energy : np.ndarray
            The energy for each of these points.

        Raises
        ------
        AssertionError
            If the system does not have a potential function defined.
        """
        assert self.has_potential_function
        points = np.array(points).reshape((-1, self.dimension))
        return self._impl.potential(points)


class _TimeDependentRhsMixin:
    def trajectory(self: SystemBase, t0, x0, length, seed=-1, n_jobs=None, return_time=False):
        r""" Simulates one or multiple trajectories depending on initial state and initial time.

        Parameters
        ----------
        t0 : array_like
            The initial time. Can be picked as single float across all test points or individually.
        x0 : array_like
            The initial condition. Must be compatible in shape to a (n_test_points, dimension)-array.
        length : int
            The length of the trajectory that is to be generated.
        seed : int, optional, default=-1
            The random seed. In case it is specified to be something else than `-1`, n_jobs must be set to `n_jobs=1`.
        n_jobs : int, optional, default=None
            Specify number of jobs according to :meth:`deeptime.util.parallel.handle_n_jobs`.
        return_time : bool, optional, default=False
            Whether to return the evaluation times too.

        Returns
        -------
        trajectory : np.ndarray
            The trajectory of shape (n_initial_conditions, n_evaluations, dimension).
            In case of just one initial condition the shape is squeezed to (n_evaluations, dimension).
        """
        n_jobs = handle_n_jobs(n_jobs)
        x0 = np.array(x0).reshape((-1, self.dimension))
        n_initial_conditions = x0.shape[0]

        t0 = np.atleast_1d(t0)
        if len(t0) == 1 and n_initial_conditions > 1:
            t0 = np.full((n_initial_conditions,), t0[0])
        tarr, traj = self._impl.trajectory(t0, x0, length, seed, n_jobs)
        if n_initial_conditions == 1:
            traj = traj[0]
        return traj if not return_time else (tarr, traj)

    def __call__(self: SystemBase, t0, test_points, seed=-1, n_jobs=None):
        r"""Evolves the provided tests points under the dynamic for n_steps and returns.

        Parameters
        ----------
        t0 : array_like
            The initial time. Can be picked as single float across all test points or individually.
        test_points : array_like
            The test points.
        seed : int, optional, default=-1
            The seed for reproducibility. In case it is set to `seed >= 0` the number of jobs needs
            to be fixed to `n_jobs=1`.
        n_jobs : int, optional, default=None
            Specify number of jobs according to :meth:`deeptime.util.parallel.handle_n_jobs`.

        Returns
        -------
        points : np.ndarray
            The evolved test points.
        """
        n_jobs = handle_n_jobs(n_jobs)
        test_points = np.array(test_points).reshape((-1, self.dimension))
        n_test_points = len(test_points)

        t0 = np.atleast_1d(t0)
        if len(t0) == 1 and n_test_points > 1:
            t0 = np.full((n_test_points,), t0[0])

        return self._impl(t0, test_points, seed, n_jobs)

    def potential(self: SystemBase, time, points):
        r""" Evaluates the system's potential function at given points in state space at a specified time.

        Parameters
        ----------
        time : float
            The evaluation time.
        points : array_like
            The points.

        Returns
        -------
        energy : np.ndarray
            The energy for each of these points.

        Raises
        ------
        AssertionError
            If the system does not have a potential function defined.
        """
        assert self.has_potential_function
        points = np.array(points).reshape((-1, self.dimension))
        return self._impl.potential(time, points)


class TimeDependentSystem(SystemBase, _TimeDependentRhsMixin):
    r""" Wraps systems with a time-dependent right-hand side defined in extension code.

    Parameters
    ----------
    system
        The implementation.
    h : float
        Integration step size.
    n_steps : int
        Number of steps between each evaluation of the state.
    props : dict or None, optional, default=None
        Additional properties of the system.
    """

    def __init__(self, system, h: float, n_steps: int, props=None):
        super().__init__(system, h, n_steps, props)


class TimeIndependentSystem(SystemBase, _TimeIndependentRhsMixin):
    r""" Wraps systems with a time-independent right-hand side defined in extension code.

    Parameters
    ----------
    system
        The implementation.
    h : float
        Integration step size.
    n_steps : int
        Number of steps between each evaluation of the state.
    props : dict or None, optional, default=None
        Additional properties of the system.
    """

    def __init__(self, system, h: float, n_steps: int, props=None):
        super().__init__(system, h, n_steps, props)


class CustomSystem(TimeIndependentSystem):
    r""" A system as yielded by :meth:`custom_sde <deeptime.data.custom_sde>`
    or :meth:`custom_ode <deeptime.data.custom_ode>`.
    """

    def trajectory(self: SystemBase, x0, length, seed=-1, **kw):
        return super().trajectory(x0, length, seed, n_jobs=1)

    def __call__(self: SystemBase, test_points, seed=-1, **kw):
        return super().__call__(test_points, seed, n_jobs=1)



import abc
from typing import Optional, List

import numpy as np
from ._hmm_bindings import output_models as _bindings

from ...base import Model


class OutputModel(Model, metaclass=abc.ABCMeta):
    r""" Output model superclass. Contains basic functionality and interfaces which output models are supposed
    to implement.

    Parameters
    ----------
    n_hidden_states : int
        Number of hidden states.
    n_observable_states : int
        Number of observable states if discrete state space, otherwise -1
    ignore_outliers : bool, optional, default=True
        By outliers observations that have zero probability given the
        current model are meant. :code:`ignore_outliers=True` means that outliers will be treated
        as if no observation was made, which is equivalent to making this
        observation with equal probability from any hidden state.
        :code:`ignore_outliers=False` means that an Exception or in the worst case an
        unhandled crash will occur if an outlier is observed.

    See Also
    --------
    DiscreteOutputModel
    GaussianOutputModel
    """

    def __init__(self, n_hidden_states: int, n_observable_states: int, ignore_outliers: bool = True):
        super().__init__()
        self._n_hidden_states = n_hidden_states
        self._n_observable_states = n_observable_states
        self._ignore_outliers = ignore_outliers

    @property
    def ignore_outliers(self) -> bool:
        r""" By outliers observations that have zero probability given the
        current model are meant. :code:`ignore_outliers=True` means that outliers will be treated
        as if no observation was made, which is equivalent to making this
        observation with equal probability from any hidden state.
        :code:`ignore_outliers=False` means that an Exception or in the worst case an
        unhandled crash will occur if an outlier is observed.
        """
        return self._ignore_outliers

    @ignore_outliers.setter
    def ignore_outliers(self, value: bool):
        self._ignore_outliers = value

    @property
    def n_hidden_states(self) -> int:
        r""" Number of hidden states. """
        return self._n_hidden_states

    @property
    def n_observable_states(self) -> int:
        r""" Number of observable states, can be -1 if not applicable (e.g., in a continous observable space). """
        return self._n_observable_states

    @abc.abstractmethod
    def to_state_probability_trajectory(self, observations: np.ndarray) -> np.ndarray:
        r"""Converts a list of observations to hidden state probabilities, i.e., for each observation :math:`o_t`,
        one obtains a vector :math:`p_t\in\mathbb{R}^{n_\mathrm{hidden}}` denoting how probable that particular
        observation is in one of the hidden states.

        Parameters
        ----------
        observations : (T, d) ndarray
            Array of observations.

        Returns
        -------
        state_probabilities : (T, n_hidden) ndarray
            State probability trajectory.
        """

    @abc.abstractmethod
    def generate_observation_trajectory(self, hidden_state_trajectory: np.ndarray) -> np.ndarray:
        r""" Generates a synthetic trajectory in observation space given a trajectory in hidden state space.

        Parameters
        ----------
        hidden_state_trajectory : (T, 1) ndarray
            Hidden state trajectory.

        Returns
        -------
        observations : (T, d) ndarray
            Observation timeseries :math:`\{o_t\}_t`, where :math:`o_t` is the observation
            associated to hidden state :math:`s_t`.
        """

    @abc.abstractmethod
    def sample(self, observations_per_state: List[np.ndarray]) -> None:
        r""" Samples a new set of parameters for this output model given a list of observations for each hidden
        state.

        Parameters
        ----------
        observations_per_state : list of ndarray
            The observations per state, i.e., :code:`len(observations_per_state)` must evaluate to the value
            of :attr:`n_hidden_states`.
        """

    @abc.abstractmethod
    def submodel(self, states: Optional[np.ndarray] = None, obs: Optional[np.ndarray] = None):
        r"""Restricts this model to a set of hidden states and observable states (if applicable).

        Parameters
        ----------
        states : ndarray, optional, default=None
            The hidden states to restrict to, per default no restriction.
        obs : ndarray, optional, default=None
            The observable states to restrict to (if applicable), per default no restriction.

        Returns
        -------
        submodel : OutputModel
            The restricted output model.
        """

    @abc.abstractmethod
    def fit(self, observations: List[np.ndarray], weights: List[np.ndarray]):
        r""" Fits the output model given the observations and weights.

        Parameters
        ----------
        observations : list of ndarray
            A list of K observation trajectories
        weights : list of ndarray
            A list of K weight matrices, each having length T_k of their corresponding observation trajectory.
            Evaluating weights[k][t,n] should yield the weight assignment from observations[k][t] to state index n.

        Returns
        -------
        self : OutputModel
            Reference to self.
        """

    @staticmethod
    def _handle_outliers(state_probability_trajectory: np.ndarray) -> None:
        r"""Takes a state probability trajectory, shape (T, n_hidden_states), and sets all probabilities which sum up to
        0 to uniform (or rather constant one, the trajectory is re-normalized later to make it row-stochastic again).
        This operation is preformed in-place.

        Parameters
        ----------
        state_probability_trajectory : (T, n_hidden) ndarray
            state probability trajectory: for each time step, the probability to be in a certain hidden state
        """
        _bindings.handle_outliers(state_probability_trajectory)


class DiscreteOutputModel(OutputModel):
    r"""HMM output probability model using discrete symbols. This corresponds to the "standard" HMM that is
    classically used in the literature.

    Parameters
    ----------
    output_probabilities : ((N, M) dtype=float32 or float64) ndarray
        Row-stochastic output probability matrix for :code:`N` hidden states and :code:`M` observable symbols.
    prior : None or type and shape of output_probabilities, optional, default=None
        Prior for the initial distribution of the HMM. Currently implements the Dirichlet prior that is
        conjugate to the Dirichlet distribution of :math:`b_i`, which is sampled from

        .. math:
            b_i \sim \prod_j b_{ij}_i^{a_{ij} + n_{ij} - 1},

        where :math:`n_{ij}` are the number of times symbol :math:`j` has been observed when the hidden trajectory
        was in state :math:`i` and :math:`a_{ij}` is the prior count. The default prior=None corresponds
        to :math:`a_{ij} = 0`. This option ensures coincidence between sample mean an MLE.
    ignore_outliers : bool, optional, default=False
        Whether to ignore outliers, see :attr:`ignore_outliers`.
    """
    def __init__(self, output_probabilities: np.ndarray, prior: Optional[np.ndarray] = None,
                 ignore_outliers: bool = False):
        if output_probabilities.ndim != 2:
            raise ValueError("Discrete output model requires two-dimensional output probability matrix!")
        if np.any(output_probabilities < 0) or not np.allclose(output_probabilities.sum(axis=-1), 1.):
            raise ValueError("Output probabilities must be row-stochastic, i.e., the row-sums must be one and all "
                             "elements must be non-negative.")
        super(DiscreteOutputModel, self).__init__(n_hidden_states=output_probabilities.shape[0],
                                                  n_observable_states=output_probabilities.shape[1],
                                                  ignore_outliers=ignore_outliers)
        if prior is None:
            prior = np.zeros_like(output_probabilities)
        if prior.shape != output_probabilities.shape:
            raise ValueError(f"Prior must have same shape as output probabilities: {prior.shape} "
                             f"!= {output_probabilities.shape}.")
        self._output_probabilities = output_probabilities
        self._prior = prior

    @property
    def prior(self) -> np.ndarray:
        r""" Prior matrix. """
        return self._prior

    @property
    def output_probabilities(self):
        r""" A row-stochastic matrix of shape (:attr:`n_hidden_states`, :attr:`n_observable_states`) describing
        the (conditional) discrete probability distribution of observable states given one hidden state.
        """
        return self._output_probabilities

    def to_state_probability_trajectory(self, observations: np.ndarray) -> np.ndarray:
        r""" Returns the output probabilities for an entire trajectory and all hidden states.

        Parameters
        ----------
        observations : ndarray((T), dtype=int)
            a discrete trajectory of length T

        Return
        ------
        p_o : ndarray (T,N)
            The probability of generating the symbol at time point t from any of the N hidden states.
        """
        state_probabilities = _bindings.discrete.to_output_probability_trajectory(observations,
                                                                                  self.output_probabilities)
        if self.ignore_outliers:
            self._handle_outliers(state_probabilities)
        return state_probabilities

    def generate_observation_trajectory(self, hidden_state_trajectory: np.ndarray) -> np.ndarray:
        return _bindings.discrete.generate_observation_trajectory(hidden_state_trajectory, self.output_probabilities)

    def fit(self, observations: List[np.ndarray], weights: List[np.ndarray]):
        # initialize output probability matrix
        self._output_probabilities.fill(0)
        # update output probability matrix (numerator)
        for obs, w in zip(observations, weights):
            _bindings.discrete.update_p_out(obs, w, self._output_probabilities)
        self.normalize()
        return self

    def normalize(self):
        r""" Normalizes output probabilities so they are row-stochastic. """
        self._output_probabilities /= np.sum(self._output_probabilities, axis=1)[:, None]

    def submodel(self, states: Optional[np.ndarray] = None, obs: Optional[np.ndarray] = None):
        if states is None:
            states = np.arange(self.output_probabilities.shape[0])
        if obs is None:
            obs = np.arange(self.output_probabilities.shape[1])
        B = np.copy(self.output_probabilities[np.ix_(states, obs)])
        B /= B.sum(axis=1)[:, None]
        if self.prior is not None:
            prior = np.copy(self.prior[np.ix_(states, obs)])
        else:
            prior = None
        return DiscreteOutputModel(B, prior, self.ignore_outliers)

    def sample(self, observations_per_state: List[np.ndarray]) -> None:
        r"""
        Sample a new set of distribution parameters given a sample of observations from the given state.
        The internal parameters are updated.

        Parameters
        ----------
        observations_per_state :  [ numpy.array with shape (N_k,) ] of length :attr:`n_hidden_states`
            observations[k] are all observations associated with hidden state k

        Examples
        --------
        Initialize output model

        >>> B = np.array([[0.5, 0.5], [0.1, 0.9]])
        >>> output_model = DiscreteOutputModel(B)

        Sample given observation

        >>> obs = [np.asarray([0, 0, 0, 1, 1, 1]),
        ...        np.asarray([1, 1, 1, 1, 1, 1])]
        >>> output_model.sample(obs)
        """
        # todo why ignore observation states w/o counts in a hidden state parameter sample?
        _bindings.discrete.sample(observations_per_state, self.output_probabilities, self.prior)


class GaussianOutputModel(OutputModel):
    r""" HMM output probability model using one-dimensional Gaussians.

    Parameters
    ----------
    n_states : int
        number of hidden states
    means : array_like, optional, default=None
        means of the output Gaussians, length must match number of hidden states
    sigmas : array_like, optional, default=None
        sigmas of the output Gaussians, length must match number of hidden states
    ignore_outliers : bool, optional, default=True
        whether to ignore outliers which could cause numerical instabilities
    """

    def __init__(self, n_states: int, means=None, sigmas=None,
                 ignore_outliers: bool = True):
        if means is None:
            means = np.zeros((n_states,))
        else:
            means = np.asarray(means)
        if sigmas is None:
            sigmas = np.zeros((n_states,))
        else:
            sigmas = np.asarray(sigmas)
        means, sigmas = means.squeeze(), sigmas.squeeze()
        if means.ndim != 1 or sigmas.ndim != 1:
            raise ValueError("Means and sigmas must be one-dimensional.")
        if means.shape[0] != n_states or sigmas.shape[0] != n_states:
            raise ValueError(f"The number of means and sigmas provided ({means.shape[0]} and {sigmas.shape[0]}, "
                             f"respectively) must match the number of output states.")
        self._means = means
        self._sigmas = sigmas

        super(GaussianOutputModel, self).__init__(n_hidden_states=n_states, n_observable_states=-1,
                                                  ignore_outliers=ignore_outliers)

    @property
    def means(self):
        r""" Mean values of Gaussian output densities. """
        return self._means

    @property
    def sigmas(self):
        r""" Standard deviations of Gaussian output densities. """
        return self._sigmas

    def to_state_probability_trajectory(self, observations: np.ndarray) -> np.ndarray:
        state_probabilities = _bindings.gaussian.to_output_probability_trajectory(observations, self.means, self.sigmas)
        if self.ignore_outliers:
            self._handle_outliers(state_probabilities)
        return state_probabilities

    def generate_observation_trajectory(self, hidden_state_trajectory: np.ndarray) -> np.ndarray:
        """ Generate synthetic observation data from a given state sequence.

        Parameters
        ----------
        hidden_state_trajectory : numpy.array with shape (T,) of int type
            s_t[t] is the hidden state sampled at time t

        Returns
        -------
        o_t : numpy.array with shape (T,) of type dtype
            o_t[t] is the observation associated with state s_t[t]

        Examples
        --------

        Generate an observation model and synthetic state trajectory.

        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(n_states=3, means=[-1, 0, +1], sigmas=[0.5, 1, 2])
        >>> s_t = np.random.randint(0, output_model.n_hidden_states, size=[nobs])

        Generate a synthetic trajectory

        >>> o_t = output_model.generate_observation_trajectory(s_t)

        """
        return _bindings.gaussian.generate_observation_trajectory(hidden_state_trajectory, self.means, self.sigmas)

    def submodel(self, states: Optional[np.ndarray] = None, obs: Optional[np.ndarray] = None):
        if states is None:
            states = np.arange(self.means.shape[0])
        return GaussianOutputModel(len(states), means=self.means[states], sigmas=self.sigmas[states],
                                   ignore_outliers=self.ignore_outliers)

    def fit(self, observations: List[np.ndarray], weights: List[np.ndarray]):
        """
        Fits the output model given the observations and weights

        Parameters
        ----------
        observations : [ ndarray(T_k,) ] with K elements
            A list of K observation trajectories, each having length T_k and d dimensions
        weights : [ ndarray(T_k,n_states) ] with K elements
            A list of K weight matrices, each having length T_k
            weights[k][t,n] is the weight assignment from observations[k][t] to state index n

        Examples
        --------

        Generate an observation model and samples from each state.

        >>> ntrajectories = 3
        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(n_states=3, means=np.array([-1, 0, +1]), sigmas=np.array([0.5, 1, 2]))
        >>> observations = [ np.random.randn(nobs) for _ in range(ntrajectories) ] # random observations
        >>> weights = [ np.random.dirichlet([2, 3, 4], size=nobs) for _ in range(ntrajectories) ] # random weights

        Update the observation model parameters my a maximum-likelihood fit. Fit returns self.

        >>> output_model = output_model.fit(observations, weights)

        """
        if self.means.dtype == np.float32:
            means, sigmas = _bindings.gaussian.fit32(self.n_hidden_states, observations, weights)
        else:
            means, sigmas = _bindings.gaussian.fit64(self.n_hidden_states, observations, weights)

        self._means = means
        self._sigmas = sigmas
        if np.any(self._sigmas < np.finfo(self._sigmas.dtype).eps):
            raise RuntimeError('at least one sigma is too small to continue.')
        return self

    def sample(self, observations_per_state: List[np.ndarray]) -> None:
        """
        Sample a new set of distribution parameters given a sample of observations from the given state.

        The internal parameters are updated accordingly.

        Parameters
        ----------
        observations_per_state :  [ numpy.array with shape (N_k,) ] with `n_states` elements
            observations[k] is a set of observations sampled from state `k`

        Examples
        --------

        Generate synthetic observations.

        >>> n_states = 3
        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(n_states=n_states, means=np.array([-1, 0, 1]),
        ...                                    sigmas=np.array([0.5, 1, 2]))
        >>> observations = [output_model.generate_observation_trajectory(np.array([state_index]*nobs))
        ...                 for state_index in range(n_states)]

        Update output parameters by sampling.

        >>> output_model.sample(observations)

        """
        for state_index in range(self.n_hidden_states):
            # Update state emission distribution parameters.

            observations_in_state = observations_per_state[state_index]
            # Determine number of samples in this state.
            nsamples_in_state = len(observations_in_state)

            # Skip update if no observations.
            if nsamples_in_state == 0:
                import warnings
                warnings.warn('Warning: State %d has no observations.' % state_index)
            if nsamples_in_state > 0:  # Sample new mu.
                self.means[state_index] = np.random.randn() * self.sigmas[state_index] / np.sqrt(
                    nsamples_in_state) + np.mean(observations_in_state)
            if nsamples_in_state > 1:  # Sample new sigma
                # This scheme uses the improper Jeffreys prior on sigma^2, P(mu, sigma^2) \propto 1/sigma
                chisquared = np.random.chisquare(nsamples_in_state - 1)
                sigmahat2 = np.mean((observations_in_state - self.means[state_index]) ** 2)
                self.sigmas[state_index] = np.sqrt(sigmahat2) / np.sqrt(chisquared / nsamples_in_state)



from typing import Optional, Union

import numpy as np
from scipy.sparse import issparse

from deeptime.util.types import ensure_dtraj_list


def _regularize_hidden(initial_distribution, transition_matrix, reversible=True, stationary=False, count_matrix=None,
                       eps=None):
    """ Regularizes the hidden initial distribution and transition matrix.

    Makes sure that the hidden initial distribution and transition matrix have
    nonzero probabilities by setting them to eps and then renormalizing.
    Avoids zeros that would cause estimation algorithms to crash or get stuck
    in suboptimal states.

    Parameters
    ----------
    initial_distribution : ndarray(n)
        Initial hidden distribution of the HMM
    transition_matrix : ndarray(n, n)
        Hidden transition matrix
    reversible : bool
        HMM is reversible. Will make sure it is still reversible after modification.
    stationary : bool
        p0 is the stationary distribution of P. In this case, will not regularize
        p0 separately. If stationary=False, the regularization will be applied to p0.
    count_matrix : ndarray(n, n)
        Hidden count matrix. Only needed for stationary=True and P disconnected.
    eps : float or None
        minimum value of the resulting transition matrix. Default: evaluates
        to 0.01 / n. The coarse-graining equation can lead to negative elements
        and thus epsilon should be set to at least 0. Positive settings of epsilon
        are similar to a prior and enforce minimum positive values for all
        transition probabilities.

    Return
    ------
    p0 : ndarray(n)
        regularized initial distribution
    P : ndarray(n, n)
        regularized transition matrix

    """
    # input
    n = transition_matrix.shape[0]
    if eps is None:  # default output probability, in order to avoid zero columns
        eps = 0.01 / n

    # REGULARIZE P
    transition_matrix = np.maximum(transition_matrix, eps)
    # and renormalize
    transition_matrix /= transition_matrix.sum(axis=1)[:, None]
    # ensure reversibility
    if reversible:
        from deeptime.markov._transition_matrix import enforce_reversible_on_closed
        transition_matrix = enforce_reversible_on_closed(transition_matrix)

    # REGULARIZE p0
    if stationary:
        from deeptime.markov._transition_matrix import stationary_distribution
        stationary_distribution(transition_matrix, C=count_matrix)
    else:
        initial_distribution = np.maximum(initial_distribution, eps)
        initial_distribution /= initial_distribution.sum()

    return initial_distribution, transition_matrix


def _regularize_pobs(output_probabilities, nonempty=None, separate=None, eps=None):
    """ Regularizes the output probabilities.

    Makes sure that the output probability distributions has
    nonzero probabilities by setting them to eps and then renormalizing.
    Avoids zeros that would cause estimation algorithms to crash or get stuck
    in suboptimal states.

    Parameters
    ----------
    output_probabilities : ndarray(n, m)
        HMM output probabilities
    nonempty : None or iterable of int
        Nonempty set. Only regularize on this subset.
    separate : None or iterable of int
        Force the given set of observed states to stay in a separate hidden state.
        The remaining n_states-1 states will be assigned by a metastable decomposition.

    Returns
    -------
    output_probabilities : ndarray(n, m)
        Regularized output probabilities

    """
    # input
    output_probabilities = output_probabilities.copy()  # modify copy
    n, m = output_probabilities.shape  # number of hidden / observable states
    if eps is None:  # default output probability, in order to avoid zero columns
        eps = 0.01 / m
    # observable sets
    if nonempty is None:
        nonempty = np.arange(m)

    if separate is None:
        output_probabilities[:, nonempty] = np.maximum(output_probabilities[:, nonempty], eps)
    else:
        nonempty_nonseparate = np.array(list(set(nonempty) - set(separate)), dtype=int)
        nonempty_separate = np.array(list(set(nonempty).intersection(set(separate))), dtype=int)
        output_probabilities[:n - 1, nonempty_nonseparate] = np.maximum(
            output_probabilities[:n - 1, nonempty_nonseparate], eps)
        output_probabilities[n - 1, nonempty_separate] = np.maximum(output_probabilities[n - 1, nonempty_separate], eps)

    # renormalize and return copy
    output_probabilities /= output_probabilities.sum(axis=1)[:, None]
    return output_probabilities


def _coarse_grain_transition_matrix(P, M):
    """ Coarse grain transition matrix P using memberships M

    Computes

    .. math::
        P_c = (M' M)^-1 M' P M

    Parameters
    ----------
    P : ndarray(n, n)
        microstate transition matrix
    M : ndarray(n, m)
        membership matrix. Membership to macrostate m for each microstate.

    Returns
    -------
    P_coarse : ndarray(m, m)
        coarse-grained transition matrix.

    """
    # coarse-grain matrix: Pc = (M' M)^-1 M' P M
    W = np.linalg.inv(np.dot(M.T, M))
    A = np.dot(np.dot(M.T, P), M)
    P_coarse = np.dot(W, A)

    # this coarse-graining can lead to negative elements. Setting them to zero here.
    P_coarse = np.maximum(P_coarse, 0)
    # and renormalize
    P_coarse /= P_coarse.sum(axis=1)[:, None]

    return P_coarse


def metastable_from_msm(msm, n_hidden_states: int,
                        reversible: bool = True, stationary: bool = False,
                        separate_symbols=None, regularize: bool = True):
    r""" Makes an initial guess for an :class:`HMM <deeptime.markov.hmm.HiddenMarkovModel>` with
    discrete output model from an already existing MSM over observable states. The procedure is described in
    :footcite:`noe2013projected` and uses PCCA+ :footcite:`roblitz2013fuzzy` for
    coarse-graining the transition matrix and obtaining membership assignments.

    Parameters
    ----------
    msm : MarkovStateModel
        The markov state model over observable state space.
    n_hidden_states : int
        The desired number of hidden states.
    reversible : bool, optional, default=True
        Whether the HMM transition matrix is estimated so that it is reversibe.
    stationary : bool, optional, default=False
        If True, the initial distribution of hidden states is self-consistently computed as the stationary
        distribution of the transition matrix. If False, it will be estimated from the starting states.
        Only set this to true if you're sure that the observation trajectories are initiated from a global
        equilibrium distribution.
    separate_symbols : array_like, optional, default=None
        Force the given set of observed states to stay in a separate hidden state.
        The remaining nstates-1 states will be assigned by a metastable decomposition.
    regularize : bool, optional, default=True
        If set to True, makes sure that the hidden initial distribution and transition matrix have nonzero probabilities
        by setting them to eps and then renormalizing. Avoids zeros that would cause estimation algorithms to crash or
        get stuck in suboptimal states.

    Returns
    -------
    hmm_init : HiddenMarkovModel
        An initial guess for the HMM

    See Also
    --------
    deeptime.markov.hmm.DiscreteOutputModel
        The type of output model this heuristic uses.

    :func:`metastable_from_data`
        Initial guess from data if no MSM is available yet.

    :func:`deeptime.markov.hmm.init.gaussian.from_data`
        Initial guess with :class:`Gaussian output model <deeptime.markov.hmm.GaussianOutputModel>`.

    References
    ----------
    .. footbibliography::
    """
    from deeptime.markov._transition_matrix import stationary_distribution
    from deeptime.markov._transition_matrix import estimate_P
    from deeptime.markov.msm import MarkovStateModel
    from deeptime.markov import PCCAModel

    count_matrix = msm.count_model.count_matrix
    nonseparate_symbols = np.arange(msm.count_model.n_states_full)
    nonseparate_states = msm.count_model.symbols_to_states(nonseparate_symbols)
    nonseparate_msm = msm
    if separate_symbols is not None:
        separate_symbols = np.asanyarray(separate_symbols)
        if np.max(separate_symbols) >= msm.count_model.n_states_full:
            raise ValueError(f'Separate set has indices that do not exist in '
                             f'full state space: {np.max(separate_symbols)}')
        nonseparate_symbols = np.setdiff1d(nonseparate_symbols, separate_symbols)
        nonseparate_states = msm.count_model.symbols_to_states(nonseparate_symbols)
        nonseparate_count_model = msm.count_model.submodel(nonseparate_states)
        # make reversible
        nonseparate_count_matrix = nonseparate_count_model.count_matrix
        if issparse(nonseparate_count_matrix):
            nonseparate_count_matrix = nonseparate_count_matrix.toarray()
        P_nonseparate = estimate_P(nonseparate_count_matrix, reversible=True)
        pi = stationary_distribution(P_nonseparate, C=nonseparate_count_matrix)
        nonseparate_msm = MarkovStateModel(P_nonseparate, stationary_distribution=pi)
    if issparse(count_matrix):
        count_matrix = count_matrix.toarray()

    # if #metastable sets == #states, we can stop here
    n_meta = n_hidden_states if separate_symbols is None else n_hidden_states - 1
    if n_meta == nonseparate_msm.n_states:
        pcca = PCCAModel(nonseparate_msm.transition_matrix, nonseparate_msm.stationary_distribution, np.eye(n_meta),
                         np.eye(n_meta))
    else:
        pcca = nonseparate_msm.pcca(n_meta)
    if separate_symbols is not None:
        separate_states = msm.count_model.symbols_to_states(separate_symbols)
        memberships = np.zeros((msm.n_states, n_hidden_states))
        memberships[nonseparate_states, :n_hidden_states - 1] = pcca.memberships
        memberships[separate_states, -1] = 1
    else:
        memberships = pcca.memberships
        separate_states = None

    hidden_transition_matrix = _coarse_grain_transition_matrix(msm.transition_matrix, memberships)
    if reversible:
        from deeptime.markov._transition_matrix import enforce_reversible_on_closed
        hidden_transition_matrix = enforce_reversible_on_closed(hidden_transition_matrix)

    hidden_counts = memberships.T.dot(count_matrix).dot(memberships)
    hidden_pi = stationary_distribution(hidden_transition_matrix, C=hidden_counts)

    output_probabilities = np.zeros((n_hidden_states, msm.count_model.n_states_full))
    # we might have lost a few symbols, reduce nonsep symbols to the ones actually represented
    nonseparate_symbols = msm.count_model.state_symbols[nonseparate_states]
    if separate_symbols is not None:
        separate_symbols = msm.count_model.state_symbols[separate_states]
        output_probabilities[:n_hidden_states - 1, nonseparate_symbols] = pcca.metastable_distributions
        output_probabilities[-1, separate_symbols] = msm.stationary_distribution[separate_states]
    else:
        output_probabilities[:, nonseparate_symbols] = pcca.metastable_distributions

    # regularize
    eps_a = 0.01 / n_hidden_states if regularize else 0.
    hidden_pi, hidden_transition_matrix = _regularize_hidden(hidden_pi, hidden_transition_matrix, reversible=reversible,
                                                             stationary=stationary, count_matrix=hidden_counts, eps=eps_a)
    eps_b = 0.01 / msm.n_states if regularize else 0.
    output_probabilities = _regularize_pobs(output_probabilities, nonempty=None, separate=separate_symbols, eps=eps_b)
    from deeptime.markov.hmm import HiddenMarkovModel
    return HiddenMarkovModel(transition_model=hidden_transition_matrix, output_model=output_probabilities,
                             initial_distribution=hidden_pi)


def metastable_from_data(dtrajs, n_hidden_states, lagtime, stride=1, mode='largest-regularized',
                         reversible: bool = True, stationary: bool = False,
                         separate_symbols=None, states: Optional[np.ndarray] = None,
                         regularize: bool = True, connectivity_threshold: Union[str, float] = 0.):
    r"""Estimates an initial guess :class:`HMM <deeptime.markov.hmm.HiddenMarkovModel>` from given
    discrete trajectories.

    Following the procedure described in :footcite:`noe2013projected`: First
    a :class:`MSM <deeptime.markov.msm.MarkovStateModel>` is estimated, which is then subsequently
    coarse-grained with PCCA+ :footcite:`roblitz2013fuzzy`. After estimation of the MSM, this
    method calls :meth:`metastable_from_msm`.

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        A discrete trajectory or a list of discrete trajectories.
    n_hidden_states : int
        Number of hidden states.
    lagtime : int
        The lagtime at which transitions are counted.
    stride : int or str, optional, default=1
        stride between two lagged trajectories extracted from the input trajectories. Given trajectory :code:`s[t]`,
        stride and lag will result in trajectories

            :code:`s[0], s[lag], s[2 lag], ...`

            :code:`s[stride], s[stride + lag], s[stride + 2 lag], ...`

        Setting stride = 1 will result in using all data (useful for maximum likelihood estimator), while a Bayesian
        estimator requires a longer stride in order to have statistically uncorrelated trajectories. Setting
        :code:`stride='effective'` uses the largest neglected timescale as an estimate for the correlation time
        and sets the stride accordingly.
    mode : str, optional, default='largest-regularized'
        The mode at which the markov state model is estimated. Since the process is assumed to be reversible and
        finite statistics might lead to unconnected regions in state space, a subselection can automatically be made
        and the count matrix can be regularized. The following options are available:

        * 'all': all available states are taken into account
        * 'largest': the largest connected state set is selected, see
          :meth:`TransitionCountModel.submodel_largest <deeptime.markov.TransitionCountModel.submodel_largest>`.
        * populus: the connected set with the largest population in the data, see
          :meth:`TransitionCountModel.submodel_largest <deeptime.markov.TransitionCountModel.submodel_largest>`.

        For regularization, each of the options can be suffixed by a '-regularized', e.g., 'largest-regularized'.
        This means that the count matrix has no zero entries and everything is reversibly connected. In particular,
        a prior of the form

        .. math:: b_{ij}=\left \{ \begin{array}{rl}
                     \alpha & \text{, if }c_{ij}+c_{ji}>0, \\
                     0      & \text{, otherwise,}
                     \end{array} \right .

        with :math:`\alpha=10^{-3}` is added and all non-reversibly connected components are artifically connected
        by adding backward paths.
    reversible : bool, optional, default=True
        Whether the HMM transition matrix is estimated so that it is reversibe.
    stationary : bool, optional, default=False
        If True, the initial distribution of hidden states is self-consistently computed as the stationary
        distribution of the transition matrix. If False, it will be estimated from the starting states.
        Only set this to true if you're sure that the observation trajectories are initiated from a global
        equilibrium distribution.
    separate_symbols : array_like, optional, default=None
        Force the given set of observed states to stay in a separate hidden state.
        The remaining nstates-1 states will be assigned by a metastable decomposition.
    states : (dtype=int) ndarray, optional, default=None
        Artifically restrict count model to selection of states, even before regularization.
    regularize : bool, optional, default=True
        If set to True, makes sure that the hidden initial distribution and transition matrix have nonzero probabilities
        by setting them to eps and then renormalizing. Avoids zeros that would cause estimation algorithms to crash or
        get stuck in suboptimal states.
    connectivity_threshold : float or '1/n', optional, default=0.
        Connectivity threshold. counts that are below the specified value are disregarded when finding connected
        sets. In case of '1/n', the threshold gets resolved to :math:`1 / \mathrm{n\_states\_full}`.

    Returns
    -------
    hmm_init : HiddenMarkovModel
        An initial guess for the HMM

    See Also
    --------
    DiscreteOutputModel
        The type of output model this heuristic uses.

    :func:`metastable_from_msm`
        Initial guess from an already existing :class:`MSM <deeptime.markov.msm.MarkovStateModel>`.

    :func:`deeptime.markov.hmm.init.gaussian.from_data`
        Initial guess with :class:`Gaussian output model <deeptime.markov.hmm.GaussianOutputModel>`.


    References
    ----------
    .. footbibliography::
    """
    if mode not in metastable_from_data.VALID_MODES \
            + [m + "-regularized" for m in metastable_from_data.VALID_MODES]:
        raise ValueError("mode can only be one of [{}]".format(", ".join(metastable_from_data.VALID_MODES)))

    from deeptime.markov import compute_dtrajs_effective, TransitionCountEstimator

    dtrajs = ensure_dtraj_list(dtrajs)
    dtrajs = compute_dtrajs_effective(dtrajs, lagtime=lagtime, n_states=n_hidden_states, stride=stride)
    counts = TransitionCountEstimator(1, 'sliding', sparse=False).fit(dtrajs).fetch_model()
    if states is not None:
        counts = counts.submodel(states)
    if '-regularized' in mode:
        import deeptime.markov.tools.estimation as memest
        counts.count_matrix[...] += memest.prior_neighbor(counts.count_matrix, 0.001)
        nonempty = np.where(counts.count_matrix.sum(axis=0) + counts.count_matrix.sum(axis=1) > 0)[0]
        counts.count_matrix[nonempty, nonempty] = np.maximum(counts.count_matrix[nonempty, nonempty], 0.001)
    if 'all' in mode:
        pass  # no-op
    if 'largest' in mode:
        counts = counts.submodel_largest(directed=True, connectivity_threshold=connectivity_threshold,
                                         sort_by_population=False)
    if 'populous' in mode:
        counts = counts.submodel_largest(directed=True, connectivity_threshold=connectivity_threshold,
                                         sort_by_population=True)
    from deeptime.markov.msm import MaximumLikelihoodMSM
    msm = MaximumLikelihoodMSM(reversible=True, allow_disconnected=True, maxerr=1e-3,
                               maxiter=10000).fit(counts).fetch_model()
    return metastable_from_msm(msm, n_hidden_states, reversible, stationary, separate_symbols, regularize)


metastable_from_data.VALID_MODES = ['all', 'largest', 'populous']


def random_guess(n_observation_states: int, n_hidden_states: int, seed: Optional[int] = None):
    r"""Initializes a :class:`HMM <deeptime.markov.hmm.HiddenMarkovModel>` with a set number of hidden and
    observable states by setting the transition matrix uniform and drawing a random row-stochastic matrix as
    output probabilities.

    Parameters
    ----------
    n_observation_states : int
        The number of states in observable space.
    n_hidden_states : int
        The number of hidden states.
    seed : int, optional, default=None
        The random seed.

    Returns
    -------
    init_hmm : HiddenMarkovModel
        A randomly initialized hidden markov state model.
    """
    state = np.random.RandomState(seed=seed)
    P = np.empty((n_hidden_states, n_hidden_states))
    P.fill(1. / n_hidden_states)
    B = state.uniform(size=(n_hidden_states, n_observation_states))
    B /= B.sum(axis=-1, keepdims=True)
    from deeptime.markov.hmm import HiddenMarkovModel
    return HiddenMarkovModel(transition_model=P, output_model=B)



"""
@author: Susanna Roeblitz, Marcus Weber, Frank Noe

modified from ZIBMolPy which can also be found on Github:
https://github.com/CMD-at-ZIB/ZIBMolPy/blob/master/ZIBMolPy_package/ZIBMolPy/algorithms.py
"""

import warnings
import numpy as np
from scipy.sparse import issparse
import math


def _pcca_connected_isa(eigenvectors, n_clusters):
    """
    PCCA+ spectral clustering method using the inner simplex algorithm.

    Clusters the first n_cluster eigenvectors of a transition matrix in order to cluster the states.
    This function assumes that the state space is fully connected, i.e. the transition matrix whose
    eigenvectors are used is supposed to have only one eigenvalue 1, and the corresponding first
    eigenvector (evec[:,0]) must be constant.

    Parameters
    ----------
    eigenvectors : ndarray
        A matrix with the sorted eigenvectors in the columns. The stationary eigenvector should
        be first, then the one to the slowest relaxation process, etc.

    n_clusters : int
        Number of clusters to group to.

    Returns
    -------
    chi : ndarray (n x m)
        A matrix containing the probability or membership of each state to be assigned to each cluster.
        The rows sum to 1.

    rot_mat : ndarray (m x m)
        A rotation matrix that rotates the dominant eigenvectors to yield the PCCA memberships, i.e.:
        chi = np.dot(evec, rot_matrix

    References
    ----------
    [1] P. Deuflhard and M. Weber, Robust Perron cluster analysis in conformation dynamics.
        in: Linear Algebra Appl. 398C M. Dellnitz and S. Kirkland and M. Neumann and C. Schuette (Editors)
        Elsevier, New York, 2005, pp. 161-184

    """
    (n, m) = eigenvectors.shape

    # do we have enough eigenvectors?
    if n_clusters > m:
        raise ValueError("Cannot cluster the (" + str(n) + " x " + str(m)
                         + " eigenvector matrix to " + str(n_clusters) + " clusters.")

    # check if the first, and only the first eigenvector is constant
    diffs = np.abs(np.max(eigenvectors, axis=0) - np.min(eigenvectors, axis=0))
    assert diffs[0] < 1e-6, "First eigenvector is not constant. This indicates that the transition matrix " \
                            "is not connected or the eigenvectors are incorrectly sorted. Cannot do PCCA."
    assert diffs[1] > 1e-6, "An eigenvector after the first one is constant. " \
                            "Probably the eigenvectors are incorrectly sorted. Cannot do PCCA."

    # local copy of the eigenvectors
    c = eigenvectors[:, list(range(n_clusters))]

    ortho_sys = np.copy(c)
    max_dist = 0.0

    # representative states
    ind = np.zeros(n_clusters, dtype=np.int32)

    # select the first representative as the most outlying point
    for (i, row) in enumerate(c):
        if np.linalg.norm(row, 2) > max_dist:
            max_dist = np.linalg.norm(row, 2)
            ind[0] = i

    # translate coordinates to make the first representative the origin
    ortho_sys -= c[ind[0], None]

    # select the other m-1 representatives using a Gram-Schmidt orthogonalization
    for k in range(1, n_clusters):
        max_dist = 0.0
        temp = np.copy(ortho_sys[ind[k - 1]])

        # select next farthest point that is not yet a representative
        for (i, row) in enumerate(ortho_sys):
            row -= np.dot(np.dot(temp, np.transpose(row)), temp)
            distt = np.linalg.norm(row, 2)
            if distt > max_dist and i not in ind[0:k]:
                max_dist = distt
                ind[k] = i
        ortho_sys /= np.linalg.norm(ortho_sys[ind[k]], 2)

    # print "Final selection ", ind

    # obtain transformation matrix of eigenvectors to membership matrix
    rot_mat = np.linalg.inv(c[ind])
    # print "Rotation matrix \n ", rot_mat

    # compute membership matrix
    chi = np.dot(c, rot_mat)
    # print "chi \n ", chi

    return chi, rot_mat


def _opt_soft(eigenvectors, rot_matrix, n_clusters):
    """
    Optimizes the PCCA+ rotation matrix such that the memberships are exclusively nonnegative.

    Parameters
    ----------
    eigenvectors : ndarray
        A matrix with the sorted eigenvectors in the columns. The stationary eigenvector should
        be first, then the one to the slowest relaxation process, etc.

    rot_matrix : ndarray (m x m)
        nonoptimized rotation matrix

    n_clusters : int
        Number of clusters to group to.

    Returns
    -------
    rot_mat : ndarray (m x m)
        Optimized rotation matrix that rotates the dominant eigenvectors to yield the PCCA memberships, i.e.:
        chi = np.dot(evec, rot_matrix

    References
    ----------
    [1] S. Roeblitz and M. Weber, Fuzzy spectral clustering by PCCA+:
        application to Markov state models and data classification.
        Adv Data Anal Classif 7, 147-179 (2013).

    """
    # only consider first n_clusters eigenvectors
    eigenvectors = eigenvectors[:, :n_clusters]

    # crop first row and first column from rot_matrix
    # rot_crop_matrix = rot_matrix[1:,1:]
    rot_crop_matrix = rot_matrix[1:][:, 1:]

    (x, y) = rot_crop_matrix.shape

    # reshape rot_crop_matrix into linear vector
    rot_crop_vec = np.reshape(rot_crop_matrix, x * y)

    # Susanna Roeblitz' target function for optimization
    def susanna_func(rot_crop_vec, eigvectors):
        # reshape into matrix
        rot_crop_matrix = np.reshape(rot_crop_vec, (x, y))
        # fill matrix
        rot_matrix = _fill_matrix(rot_crop_matrix, eigvectors)

        result = 0
        for i in range(0, n_clusters):
            for j in range(0, n_clusters):
                result += np.power(rot_matrix[j, i], 2) / rot_matrix[0, i]
        return -result

    from scipy.optimize import fmin

    rot_crop_vec_opt = fmin(susanna_func, rot_crop_vec, args=(eigenvectors,), disp=False)

    rot_crop_matrix = np.reshape(rot_crop_vec_opt, (x, y))
    rot_matrix = _fill_matrix(rot_crop_matrix, eigenvectors)

    return rot_matrix


def _fill_matrix(rot_crop_matrix, eigvectors):
    """ Helper function for opt_soft """

    (x, y) = rot_crop_matrix.shape

    row_sums = np.sum(rot_crop_matrix, axis=1)
    row_sums = np.reshape(row_sums, (x, 1))

    # add -row_sums as leftmost column to rot_crop_matrix
    rot_crop_matrix = np.concatenate((-row_sums, rot_crop_matrix), axis=1)

    tmp = -np.dot(eigvectors[:, 1:], rot_crop_matrix)

    tmp_col_max = np.max(tmp, axis=0)
    tmp_col_max = np.reshape(tmp_col_max, (1, y + 1))

    tmp_col_max_sum = np.sum(tmp_col_max)

    # add col_max as top row to rot_crop_matrix and normalize
    rot_matrix = np.concatenate((tmp_col_max, rot_crop_matrix), axis=0)
    rot_matrix /= tmp_col_max_sum

    return rot_matrix


def _pcca_connected(P, n, pi=None):
    r"""PCCA+ spectral clustering method with optimized memberships [1]_

    Clusters the first n_cluster eigenvectors of a transition matrix in order to cluster the states.
    This function assumes that the transition matrix is fully connected.

    Parameters
    ----------
    P : ndarray (n,n)
        Transition matrix.
    n : int
        Number of clusters to group to.
    pi: ndarray(n,), optional, default=None
        Stationary distribution if available.

    Returns
    -------
    chi : ndarray (n x m)
        A matrix containing the probability or membership of each state to be assigned to each cluster.
        The rows sum to 1.

    References
    ----------
    [1] S. Roeblitz and M. Weber, Fuzzy spectral clustering by PCCA+:
        application to Markov state models and data classification.
        Adv Data Anal Classif 7, 147-179 (2013).
    """

    # test connectivity
    from deeptime.markov.tools.estimation import connected_sets

    labels = connected_sets(P)
    n_components = len(labels)  # (n_components, labels) = connected_components(P, connection='strong')
    if n_components > 1:
        raise ValueError("Transition matrix is disconnected. Cannot use pcca_connected.")

    if pi is None:
        from deeptime.markov.tools.analysis import stationary_distribution
        pi = stationary_distribution(P)
    else:
        if pi.shape[0] != P.shape[0]:
            raise ValueError(f"Stationary distribution must span entire state space but got {pi.shape[0]} states "
                             f"instead of {P.shape[0]}.")
        pi /= pi.sum()  # make sure it is normalized

    from deeptime.markov.tools.analysis import is_reversible

    if not is_reversible(P, mu=pi):
        raise ValueError("Transition matrix does not fulfill detailed balance. "
                         "Make sure to call pcca with a reversible transition matrix estimate")
    # TODO: Susanna mentioned that she has a potential fix for nonreversible matrices by replacing each complex conjugate
    #      pair by the real and imaginary components of one of the two vectors. We could use this but would then need to
    #      orthonormalize all eigenvectors e.g. using Gram-Schmidt orthonormalization. Currently there is no theoretical
    #      foundation for this, so I'll skip it for now.

    # right eigenvectors, ordered
    from deeptime.markov.tools.analysis import eigenvectors

    evecs = eigenvectors(P, n)

    # orthonormalize
    for i in range(n):
        evecs[:, i] /= math.sqrt(np.dot(evecs[:, i] * pi, evecs[:, i]))
    # make first eigenvector positive
    evecs[:, 0] = np.abs(evecs[:, 0])

    # Is there a significant complex component?
    if not np.alltrue(np.isreal(evecs)):
        warnings.warn(
            "The given transition matrix has complex eigenvectors, so it doesn't exactly fulfill detailed balance. "
            "Forcing eigenvectors to be real and continuing. Be aware that this is not theoretically solid.")
    evecs = np.real(evecs)

    # create initial solution using PCCA+. This could have negative memberships
    chi, rot_matrix = _pcca_connected_isa(evecs, n)

    # optimize the rotation matrix with PCCA++.
    rot_matrix = _opt_soft(evecs, rot_matrix, n)

    # These memberships should be nonnegative
    memberships = np.dot(evecs[:, :], rot_matrix)

    # We might still have numerical errors. Force memberships to be in [0,1]
    memberships = np.clip(memberships, 0., 1.)

    for i in range(0, np.shape(memberships)[0]):
        memberships[i] /= np.sum(memberships[i])

    return memberships


def pcca(P: np.ndarray, m: int, pi: np.ndarray = None, transition_matrix_tol: float = 1e-12):
    """
    PCCA+ spectral clustering method with optimized memberships [1]_

    Clusters the first m eigenvectors of a transition matrix in order to cluster the states.
    This function does not assume that the transition matrix is fully connected. Disconnected sets
    will automatically define the first metastable states, with perfect membership assignments.

    Parameters
    ----------
    P : ndarray (n,n)
        Transition matrix.
    m : int
        Number of clusters to group to.
    pi : ndarray(n,), optional, default=None
        Stationary distribution if available. Should be defined piecewise over the connected sets.
    transition_matrix_tol : float, optional, default=1e-12
        Tolerance under which P is checked to be a transition matrix.

    Returns
    -------
    chi : ndarray (n x m)
        A matrix containing the probability or membership of each state to be assigned to each cluster.
        The rows sum to 1.

    References
    ----------
    [1] S. Roeblitz and M. Weber, Fuzzy spectral clustering by PCCA+:
        application to Markov state models and data classification.
        Adv Data Anal Classif 7, 147-179 (2013).
    [2] F. Noe, multiset PCCA and HMMs, in preparation.
    """
    # imports
    from deeptime.markov.tools.estimation import connected_sets
    from deeptime.markov.tools.analysis import eigenvalues, is_transition_matrix, hitting_probability

    # validate input
    n = np.shape(P)[0]
    if m > n:
        raise ValueError(f"Number of metastable states m={m} exceeds number of states of transition matrix n={n}")
    if not is_transition_matrix(P, tol=transition_matrix_tol):
        raise ValueError("Input matrix is not a transition matrix.")
    if pi is not None and pi.ndim > 1:
        raise ValueError("Stationary distribution must be given as one-dimensional array or left None.")
    if pi is not None and pi.shape[0] != n:
        raise ValueError(f"Stationary distribution must be defined on entire space, piecewise if the transition matrix "
                         f"has multiple connected components. It covered {pi.shape[0]} != {n} states.")

    # prepare output
    chi = np.zeros((n, m))

    # test connectivity
    components = connected_sets(P)
    n_components = len(components)  # (n_components, labels) = connected_components(P, connection='strong')

    # store components as closed (with positive equilibrium distribution)
    # or as transition states (with vanishing equilibrium distribution)
    closed_components = []
    transition_states = []
    for i in range(n_components):
        component = components[i]  # np.argwhere(labels==i).flatten()
        rest = list(set(range(n)) - set(component))
        # is component closed?
        if np.sum(P[component, :][:, rest]) == 0:
            closed_components.append(component)
        else:
            transition_states.append(component)
    n_closed_components = len(closed_components)
    closed_states = np.concatenate(closed_components)
    if len(transition_states) == 0:
        transition_states = np.array([], dtype=int)
    else:
        transition_states = np.concatenate(transition_states)

    # check if we have enough clusters to support the disconnected sets
    if m < len(closed_components):
        raise ValueError(f"Number of metastable states m={m} is too small. Transition matrix "
                         f"has {len(closed_components)} disconnected components.")

    # We collect eigenvalues in order to decide which
    closed_components_Psub = []
    closed_components_ev = []
    closed_components_enum = []
    for i in range(n_closed_components):
        component = closed_components[i]

        # compute eigenvalues in submatrix
        Psub = P[component, :][:, component]
        closed_components_Psub.append(Psub)
        closed_components_ev.append(eigenvalues(Psub))
        closed_components_enum.append(i * np.ones((component.size,), dtype=int))

    # flatten
    closed_components_ev_flat = np.hstack(closed_components_ev)
    closed_components_enum_flat = np.hstack(closed_components_enum)
    # which components should be clustered?
    component_indexes = closed_components_enum_flat[np.argsort(closed_components_ev_flat)][0:m]
    # cluster each component
    ipcca = 0
    for i in range(n_closed_components):
        component = closed_components[i]
        # how many PCCA states in this component?
        m_by_component = np.shape(np.argwhere(component_indexes == i))[0]

        # if 1, then the result is trivial
        if m_by_component == 1:
            chi[component, ipcca] = 1.0
            ipcca += 1
        elif m_by_component > 1:
            # print "submatrix: ",closed_components_Psub[i]
            chi[component, ipcca:ipcca + m_by_component] = _pcca_connected(
                closed_components_Psub[i], m_by_component, pi=None if pi is None else pi[closed_components[i]]
            )
            ipcca += m_by_component
        else:
            raise RuntimeError(f"Component {i} spuriously has {m_by_component} pcca sets")

    # finally assign all transition states
    if transition_states.size > 0:
        # make all closed states absorbing, so we can see which closed state we hit first
        Pabs = P.copy()
        Pabs[closed_states, :] = 0.0
        Pabs[closed_states, closed_states] = 1.0
        for i in range(closed_states.size):
            # hitting probability to each closed state
            h = hitting_probability(Pabs, closed_states[i])
            for j in range(transition_states.size):
                # transition states belong to closed states with the hitting probability, and inherit their chi
                chi[transition_states[j]] += h[transition_states[j]] * chi[closed_states[i]]

    # check if we have m metastable sets. If less than m, we must raise
    nmeta = np.count_nonzero(chi.sum(axis=0))
    if nmeta < m:
        raise RuntimeError(f"{m} metastable states requested, but transition matrix only has {nmeta}. "
                           f"Consider using a prior or request less metastable states.")

    return chi


def coarsegrain(P, n):
    """
    Coarse-grains transition matrix P to n sets using PCCA

    Coarse-grains transition matrix P such that the dominant eigenvalues are preserved, using:

    ..math:
        \tilde{P} = M^T P M (M^T M)^{-1}

    See [2]_ for the derivation of this form from the coarse-graining method first derived in [1]_.

    References
    ----------
    [1] S. Kube and M. Weber
        A coarse graining method for the identification of transition rates between molecular conformations.
        J. Chem. Phys. 126, 024103 (2007)
    [2] F. Noe, H. Wu, J.-H. Prinz and N. Plattner:
        Projected and hidden Markov models for calculating kinetics and metastable states of complex molecules
        J. Chem. Phys. 139, 184114 (2013)
    """
    M = pcca(P, n)
    # coarse-grained transition matrix
    W = np.linalg.inv(np.dot(M.T, M))
    A = np.dot(np.dot(M.T, P), M)
    P_coarse = np.dot(W, A)

    # symmetrize and renormalize to eliminate numerical errors
    from deeptime.markov.tools.analysis import stationary_distribution
    pi_coarse = np.dot(M.T, stationary_distribution(P))
    X = np.dot(np.diag(pi_coarse), P_coarse)
    P_coarse = X / X.sum(axis=1)[:, None]

    return P_coarse


class PCCA:
    """
    PCCA+ spectral clustering method with optimized memberships [1]_

    Clusters the first m eigenvectors of a transition matrix in order to cluster the states.
    This function does not assume that the transition matrix is fully connected. Disconnected sets
    will automatically define the first metastable states, with perfect membership assignments.

    Parameters
    ----------
    P : ndarray (n,n)
        Transition matrix.
    m : int
        Number of clusters to group to.

    References
    ----------
    [1] S. Roeblitz and M. Weber, Fuzzy spectral clustering by PCCA+:
        application to Markov state models and data classification.
        Adv Data Anal Classif 7, 147-179 (2013).
    [2] F. Noe, multiset PCCA and HMMs, in preparation.
    [3] F. Noe, H. Wu, J.-H. Prinz and N. Plattner:
        Projected and hidden Markov models for calculating kinetics and metastable states of complex molecules
        J. Chem. Phys. 139, 184114 (2013)

    """

    def __init__(self, P, m):
        # TODO: can be improved: if we have eigendecomposition already, this can be exploited.
        # remember input
        if issparse(P):
            warnings.warn('pcca is only implemented for dense matrices, '
                          'converting sparse transition matrix to dense ndarray.')
            P = P.toarray()
        self.P = P
        self.m = m

        # pcca coarse-graining
        # --------------------
        # PCCA memberships
        # TODO: can be improved. pcca computes stationary distribution internally, we don't need to compute it twice.
        self._M = pcca(P, m)

        # stationary distribution
        from deeptime.markov.tools.analysis import stationary_distribution as _sd

        self._pi = _sd(P)

        # coarse-grained stationary distribution
        self._pi_coarse = np.dot(self._M.T, self._pi)

        # HMM output matrix
        self._B = np.dot(np.dot(np.diag(1.0 / self._pi_coarse), self._M.T), np.diag(self._pi))
        # renormalize B to make it row-stochastic
        self._B /= self._B.sum(axis=1)[:, None]
        self._B /= self._B.sum(axis=1)[:, None]

        # coarse-grained transition matrix
        W = np.linalg.inv(np.dot(self._M.T, self._M))
        A = np.dot(np.dot(self._M.T, P), self._M)
        self._P_coarse = np.dot(W, A)

        # symmetrize and renormalize to eliminate numerical errors
        X = np.dot(np.diag(self._pi_coarse), self._P_coarse)
        self._P_coarse = X / X.sum(axis=1)[:, None]

    @property
    def transition_matrix(self):
        return self.P

    @property
    def stationary_probability(self):
        return self._pi

    @property
    def n_metastable(self):
        return self.m

    @property
    def memberships(self):
        return self._M

    @property
    def output_probabilities(self):
        return self._B

    @property
    def coarse_grained_transition_matrix(self):
        return self._P_coarse

    @property
    def coarse_grained_stationary_probability(self):
        return self._pi_coarse

    @property
    def metastable_assignment(self):
        """
        Crisp clustering using PCCA. This is only recommended for visualization purposes. You *cannot* compute any
        actual quantity of the coarse-grained kinetics without employing the fuzzy memberships!

        Returns
        -------
        For each microstate, the metastable state it is located in.

        """
        return np.argmax(self.memberships, axis=1)

    @property
    def metastable_sets(self):
        """
        Crisp clustering using PCCA. This is only recommended for visualization purposes. You *cannot* compute any
        actual quantity of the coarse-grained kinetics without employing the fuzzy memberships!

        Returns
        -------
        A list of length equal to metastable states. Each element is an array with microstate indexes contained in it

        """
        res = []
        assignment = self.metastable_assignment
        for i in range(self.m):
            res.append(np.where(assignment == i)[0])
        return res



__author__ = 'Simon Olsson, clonker'

import logging
from typing import Optional

import numpy as np
from deeptime.markov.tools import estimation as msmest
from scipy.sparse import issparse

from deeptime.markov import TransitionCountModel
from deeptime.markov.msm import MarkovStateModel
from .._base import _MSMBaseEstimator
from ...util.stats import confidence_interval
from ...util.types import ensure_dtraj_list, ensure_traj_list


class AMMOptimizerState:
    r""" State of the optimization in AMMs. Is later attached to the MSM object for evaluation purposes. """

    def __init__(self, expectations_by_state, experimental_measurements, experimental_measurement_weights,
                 stationary_distribution, tensor_cache_size, symmetrized_count_matrix, count_matrix_row_sums):
        r""" Creates a new AMM optimizer state.

        Parameters
        ----------
        expectations_by_state : (N, K) ndarray
            See :meth:`AugmentedMSMEstimator.__init__`.
        experimental_measurements : (K,) ndarray
            See :meth:`AugmentedMSMEstimator.__init__`.
        experimental_measurement_weights : (K,) ndarray
            See :meth:`AugmentedMSMEstimator.__init__`.
        stationary_distribution : (N, N) ndarray
            Reversibly estimated stationary distribution from count matrix.
        tensor_cache_size : int
            Number of slices of the R tensor simultaneously stored. This value is computed from the cache size
            set in :meth:`AugmentedMSMEstimator.__init__`.
        symmetrized_count_matrix : (N, N) ndarray
            Symmetrized count matrix :math:`C_\mathrm{sym} = 0.5 (C + C^\top)` .
        count_matrix_row_sums : (N, ) ndarray
            Row sums of count matrix.
        """
        self.lagrange = np.zeros_like(experimental_measurements)
        self.pi = stationary_distribution
        self.pi_hat = np.copy(stationary_distribution)
        self.m_hat = None
        self.experimental_measurements = experimental_measurements
        self.experimental_measurement_weights = experimental_measurement_weights
        self.expectations_by_state = expectations_by_state
        self.tensor_cache_size = tensor_cache_size
        self.symmetrized_count_matrix = symmetrized_count_matrix
        self.count_matrix_row_sums = count_matrix_row_sums
        self.X = np.empty_like(symmetrized_count_matrix)

        self.m_hat = np.empty(())
        self._slope_obs = None
        self.update_m_hat()
        self.delta_m_hat = 1e-1 * np.ones_like(self.m_hat)
        self.R_slices = np.empty(())
        self.R_slices_i = 0
        self.update_R_slices(0)
        self.Q = np.zeros((self.n_states, self.n_states), dtype=expectations_by_state.dtype)
        self.G = np.empty(())
        self._log_likelihood_prev = None
        self.log_likelihoods = []

    @property
    def log_likelihood_prev(self):
        r""" Property storing the previous log likelihood. Used in optimization procedure.

        :getter: gets the previous log likelihood
        :setter: sets the previous log likelihood
        :type: float
        """
        return self._log_likelihood_prev

    @log_likelihood_prev.setter
    def log_likelihood_prev(self, value):
        self._log_likelihood_prev = value

    @property
    def n_states(self):
        r""" Number of discrete states (N) that fall into the considered state set. """
        return self.expectations_by_state.shape[0]

    @property
    def n_experimental_observables(self):
        r""" Number of experimental observables (K). """
        return self.expectations_by_state.shape[1]

    @property
    def slope_obs(self) -> np.ndarray:
        r""" Numerically computed slope for Newton algorithm. """
        return self._slope_obs

    def update_m_hat(self):
        r""" Updates m_hat (expectation of observable of the Augmented Markov model) """
        self.m_hat = self.pi_hat.dot(self.expectations_by_state)
        self._slope_obs = self.m_hat - self.experimental_measurements

    def update_R_slices(self, i):
        """ Computation of multiple slices of R tensor.

        When estimate(.) is called the R-tensor is split into segments whose maximum size is
        specified by max_cache argument (see constructor).
        R_slices_i specifies which of the segments are currently in cache. For equations check SI of [1].

        """
        pek = self.pi_hat[:, None] * self.expectations_by_state[:, i * self.tensor_cache_size:(i + 1) * self.tensor_cache_size]
        pp = self.pi_hat[:, None] + self.pi_hat[None, :]
        ppmhat = pp * self.m_hat[i * self.tensor_cache_size:(i + 1) * self.tensor_cache_size, None, None]
        self.R_slices = (pek[:, None, :] + pek[None, :, :]).T - ppmhat
        self.R_slices_i = i

    def update_pi_hat(self):
        r""" Update stationary distribution estimate of Augmented Markov model ( :math:`\hat pi` ) """
        expons = np.einsum('i,ji->j', self.lagrange, self.expectations_by_state)
        # expons = (self.lagrange[:, None]*self.E_active.T).sum(axis=0)
        expons = expons - np.max(expons)

        _ph_unnom = self.pi * np.exp(expons)
        self.pi_hat = (_ph_unnom / _ph_unnom.sum()).reshape(-1, )

    def log_likelihood_biased(self, count_matrix, transition_matrix):
        """ Evaluate AMM likelihood. """
        ll_unbiased = msmest.log_likelihood(count_matrix, transition_matrix)
        ll_bias = -np.sum(self.experimental_measurement_weights * (self.m_hat - self.experimental_measurements) ** 2.)
        return ll_unbiased + ll_bias

    def _get_Rk(self, k):
        """ Convenience function to get cached value of an Rk slice of the R tensor.
        If we are outside cache, update the cache and return appropriate slice.

        Parameters
        ----------
        k : int
            k-th slice.

        Returns
        -------
            The k-th slice of the R tensor.
        """
        if k > (self.R_slices_i + 1) * self.tensor_cache_size or k < self.R_slices_i * self.tensor_cache_size:
            self.update_R_slices(np.floor(k / self.tensor_cache_size).astype(int))
            return self.R_slices[k % self.tensor_cache_size]
        else:
            return self.R_slices[k % self.tensor_cache_size]

    def update_Q(self):
        """ Compute Q, a weighted sum of the R-tensor.

        See SI of [1].
        """
        self.Q.fill(0.)
        for k in range(self.n_experimental_observables):
            self.Q = self.Q + self.experimental_measurement_weights[k] * self.slope_obs[k] * self._get_Rk(k)
        self.Q *= -2.

    def update_X_and_pi(self):
        r""" Updates estimate X and stationary distribution. """
        # evaluate count-over-pi
        c_over_pi = self.count_matrix_row_sums / self.pi
        D = c_over_pi[:, None] + c_over_pi + self.Q
        # update estimate
        self.X[:] = self.symmetrized_count_matrix / D

        # renormalize
        self.X /= np.sum(self.X)
        self.pi = np.sum(self.X, axis=1)

    def update_G(self):
        """ Update G, the observable covariance.

        See SI of [1].
        """
        self.G = (np.dot(self.expectations_by_state.T, self.expectations_by_state * self.pi_hat[:, None]) -
                  self.m_hat[:, None] * self.m_hat[None, :])


class AugmentedMSM(MarkovStateModel):
    r""" An augmented Markov state model.

    Implementation following :footcite:`olsson2017combining`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, transition_matrix, stationary_distribution=None, reversible=None, n_eigenvalues=None, ncv=None,
                 count_model=None, amm_optimizer_state=None):
        super().__init__(transition_matrix=transition_matrix, stationary_distribution=stationary_distribution,
                         reversible=reversible, n_eigenvalues=n_eigenvalues, ncv=ncv, count_model=count_model)
        self._amm_optimizer_state = amm_optimizer_state
        # scoring and hmm coarse-graining not supported for augmented MSMs
        # it is recommended to set unsupported methods to None rather than raising, cf
        # https://docs.python.org/3/library/exceptions.html
        self.score = None
        self.hmm = None

    @property
    def optimizer_state(self):
        r""" The optimizer state after optimization. Can be None if instantiated directly. """
        return self._amm_optimizer_state


class AugmentedMSMEstimator(_MSMBaseEstimator):
    r""" Estimator for augmented Markov state models. :footcite:`olsson2017combining`
    This estimator is based on expectation values from experiments.
    In case the experimental data is a time series matching a discrete time series, a convenience function
    :meth:`estimator_from_feature_trajectories` is offered.

    Parameters
    ----------
    expectations_by_state : (n, k) ndarray
        Expectations by state. n Markov states, k experimental observables; each index is
        average over members of the Markov state.
    experimental_measurements : (k,) ndarray
        The experimental measurements.
    experimental_measurement_weights : (k,) ndarray
        Experimental measurements weights.
    eps : float, optional, default=0.05
        Additional convergence criterion used when some experimental data
        are outside the support of the simulation. The value of the eps
        parameter is the threshold of the relative change in the predicted
        observables as a function of fixed-point iteration:

        .. math::
            \mathrm{eps} > \frac{\mid o_{\mathrm{pred}}^{(i+1)}-o_{\mathrm{pred}}^{(i)}\mid }{\sigma}.

    support_ci : float, optional, default=1.0
        Confidence interval for determination whether experimental data are inside or outside Markov model support.
    maxiter : int, optional, default=500
        Optional parameter with specifies the maximum number of updates for Lagrange multiplier estimation.
    max_cache : int, optional, default=3000
        Maximum size (in megabytes) of cache when computing R tensor.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, expectations_by_state, experimental_measurements, experimental_measurement_weights,
                 eps=0.05, support_ci=1.00, maxiter=500, max_cache=3000):
        super().__init__(sparse=False, reversible=True)
        self.expectations_by_state = expectations_by_state
        self.experimental_measurements = experimental_measurements
        self.experimental_measurement_weights = experimental_measurement_weights
        self.convergence_criterion_lagrange = eps
        self.support_confidence = support_ci
        self.max_cache = max_cache
        self.maxiter = maxiter
        self._log = logging.getLogger(__name__)

    @staticmethod
    def estimator_from_feature_trajectories(discrete_trajectories, feature_trajectories, n_states,
                                            experimental_measurements, sigmas, eps=0.05, support_ci=1.0,
                                            maxiter=500, max_cache=3000):
        r""" Creates an AMM estimator from discrete trajectories and corresponding experimental data.
        
        Parameters
        ----------
        discrete_trajectories : array_like or list of array_like
            Discrete trajectories, stored as integer ndarrays (arbitrary size) or a single ndarray for
            only one trajectory.
        feature_trajectories : array_like or list of array_like
            The same shape (number of trajectories and timesteps) as dtrajs. Each timestep in each trajectory should
            match the shape of the measurements and sigmas, K.
        n_states : int
            Number of markov states in full state space.
        experimental_measurements : (K,) ndarray
            Experimental averages.
        sigmas : (K,) ndarray
            Standard error for each experimental observable.
        eps : float, default = 0.05
            Convergence criterium, see :meth:`__init__`.
        support_ci : float, default=1.0
            Confidence interval, see :meth:`__init__`.
        maxiter : int, optional, default=500
            Maximum number of iterations.
        max_cache : int, optional, default=3000
            Parameter which specifies the maximum size of cache used
            when performing estimation of AMM, in megabytes.

        Returns
        -------
        estimator : AugmentedMSMEstimator
            An estimator parameterized expectations by state based on feature trajectories.
        """
        discrete_trajectories = ensure_dtraj_list(discrete_trajectories)
        feature_trajectories = ensure_traj_list(feature_trajectories)
        # check input
        if np.all(sigmas > 0):
            _w = 1. / (2 * sigmas ** 2.)
        else:
            raise ValueError('Zero or negative standard errors supplied. Please revise input')
        if feature_trajectories[0].ndim < 2:
            raise ValueError("Supplied feature trajectories have inappropriate dimensions (%d), "
                             "should be at least 2." % feature_trajectories[0].ndim)
        if len(discrete_trajectories) != len(feature_trajectories):
            raise ValueError("A different number of dtrajs and ftrajs were supplied as input. "
                             "They must have exactly a one-to-one correspondence.")
        if not np.all([len(dt) == len(ft) for dt, ft in zip(discrete_trajectories, feature_trajectories)]):
            raise ValueError("One or more supplied dtraj-ftraj pairs do not have the same length.")
        # MAKE E matrix
        fta = np.concatenate(feature_trajectories)
        dta = np.concatenate(discrete_trajectories)
        _E = np.zeros((n_states, fta.shape[1]))
        for i, s in enumerate(range(n_states)):
            indices = np.where(dta == s)
            if len(indices[0]) > 0:
                _E[i, :] = fta[indices].mean(axis=0)
        # transition matrix estimator
        return AugmentedMSMEstimator(expectations_by_state=_E, experimental_measurements=experimental_measurements,
                                     experimental_measurement_weights=_w, eps=eps,
                                     support_ci=support_ci, maxiter=maxiter, max_cache=max_cache)

    @property
    def expectations_by_state(self):
        r""" The expectations by state (N) for each observable (K).

        :type: (N, K) ndarray
        """
        return self._E

    @expectations_by_state.setter
    def expectations_by_state(self, value):
        self._E = value

    @property
    def experimental_measurements(self):
        r""" Experimental measurement averages (K).

        :type: (K,) ndarray
        """
        return self._m

    @experimental_measurements.setter
    def experimental_measurements(self, value):
        self._m = value

    @property
    def experimental_measurement_weights(self):
        r""" Weights for experimental measurement averages (K).

        :type: (K,) ndarray
        """
        return self._w

    @experimental_measurement_weights.setter
    def experimental_measurement_weights(self, value):
        if np.any(value < 1e-12):
            raise ValueError("Some weights are close to zero or negative, but only weights greater or equal 1e-12 can"
                             "be dealt with appropriately.")
        self._w = value

    @property
    def uncertainties(self):
        r""" Uncertainties based on measurement weights. """
        if self.experimental_measurement_weights is not None:
            return np.sqrt(1. / 2. / self.w)
        else:
            return None

    @property
    def convergence_criterion_lagrange(self):
        r""" Additional convergence criterion used when some experimental data
        are outside the support of the simulation. The value of the eps
        parameter is the threshold of the relative change in the predicted
        observables as a function of fixed-point iteration:

        $$ \mathrm{eps} > \frac{\mid o_{\mathrm{pred}}^{(i+1)}-o_{\mathrm{pred}}^{(i)}\mid }{\sigma}. $$
        """
        return self._eps

    @convergence_criterion_lagrange.setter
    def convergence_criterion_lagrange(self, value):
        self._eps = value

    @property
    def support_confidence(self):
        r""" Confidence interval size for markov states. """
        return self._support_ci

    @support_confidence.setter
    def support_confidence(self, value):
        self._support_ci = value

    @property
    def max_cache(self):
        r""" Cache size during computation. """
        return self._max_cache

    @max_cache.setter
    def max_cache(self, value):
        self._max_cache = value

    @property
    def maxiter(self):
        r""" Maximum number of Newton iterations. """
        return self._maxiter

    @maxiter.setter
    def maxiter(self, value):
        self._maxiter = value

    def _newton_lagrange(self, state: AMMOptimizerState, count_matrix):
        """ This function performs a Newton update of the Lagrange multipliers.
        The iteration is constrained by strictly improving the AMM likelihood, and yielding meaningful stationary
        properties.
        """
        # initialize a number of values
        l_old = state.lagrange.copy()
        _ll_new = -np.inf
        frac = 1.
        mhat_old = state.m_hat.copy()
        while state.log_likelihood_prev > _ll_new or np.any(state.pi_hat < 1e-12):
            state.update_pi_hat()
            state.update_G()
            # Lagrange slope calculation
            dl = 2. * (frac * state.G * state.experimental_measurement_weights[:, None]
                       * state.slope_obs[:, None]).sum(axis=0)
            # update Lagrange multipliers
            state.lagrange = l_old - frac * dl
            state.update_pi_hat()
            # a number of sanity checks
            while np.any(state.pi_hat < 1e-12) and frac > 0.05:
                frac *= 0.5
                state.lagrange = l_old - frac * dl
                state.update_pi_hat()

            state.lagrange = l_old - frac * dl
            state.update_pi_hat()
            state.update_m_hat()
            state.update_Q()
            state.update_X_and_pi()

            transition_matrix = state.X / state.pi[:, None]
            _ll_new = state.log_likelihood_biased(count_matrix, transition_matrix)
            # decrease slope in Lagrange space (only used if loop is repeated, e.g. if sanity checks fail)
            frac *= 0.1

            if frac < 1e-12:
                self._log.info("Small gradient fraction")
                break

            state.delta_m_hat = state.m_hat - mhat_old
            state.log_likelihood_prev = float(_ll_new)

        state.log_likelihoods.append(_ll_new)

    def fetch_model(self) -> Optional[AugmentedMSM]:
        r""" Yields the most recently estimated AMM or None if :meth:`fit` was not called yet.

        Returns
        -------
        amm : AugmentedMSM or None
            The AMM instance.
        """
        return self._model

    def fit(self, data, *args, **kw):
        r""" Fits an AMM.

        Parameters
        ----------
        data : TransitionCountModel or (N, N) ndarray
            Count matrix over data.
        *args
            scikit-learn compatibility argument
        **kw
            scikit-learn compatibility argument

        Returns
        -------
        self : AugmentedMSMEstimator
            Reference to self.
        """
        if not isinstance(data, (TransitionCountModel, np.ndarray)):
            raise ValueError("Can only fit on a TransitionCountModel or a count matrix directly.")

        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[0] != data.shape[1] or np.any(data < 0.):
                raise ValueError("If fitting a count matrix directly, only non-negative square matrices can be used.")
            count_model = TransitionCountModel(data)
        else:
            count_model = data

        if len(self.experimental_measurement_weights) != self.expectations_by_state.shape[1]:
            raise ValueError("Experimental weights must span full observable space.")
        if len(self.experimental_measurements) != self.expectations_by_state.shape[1]:
            raise ValueError("Experimental measurements must span full observable state space.")

        count_matrix = count_model.count_matrix
        if issparse(count_matrix):
            count_matrix = count_matrix.toarray()

        # slice out active states from E matrix
        expectations_selected = self.expectations_by_state[count_model.state_symbols]
        count_matrix_symmetric = 0.5 * (count_matrix + count_matrix.T)
        nonzero_counts = np.nonzero(count_matrix_symmetric)
        counts_row_sums = np.sum(count_matrix, axis=1)
        expectations_confidence_interval = confidence_interval(expectations_selected, conf=self.support_confidence)

        measurements = self.experimental_measurements
        measurement_weights = self.experimental_measurement_weights

        count_outside = []
        count_inside = []

        i = 0
        # Determine which experimental values are outside the support as defined by the Confidence interval
        for confidence_lower, confidence_upper, measurement, weight in zip(
                expectations_confidence_interval[0], expectations_confidence_interval[1],
                measurements, measurement_weights):
            if measurement < confidence_lower or confidence_upper < measurement:
                self._log.info(f"Experimental value {measurement} is outside the "
                               f"support ({confidence_lower, confidence_upper})")
                count_outside.append(i)
            else:
                count_inside.append(i)
            i = i + 1

        # A number of initializations
        transition_matrix, stationary_distribution = msmest.transition_matrix(count_matrix, reversible=True,
                                                                              return_statdist=True)
        if issparse(transition_matrix):
            transition_matrix = transition_matrix.toarray()
        # Determine number of slices of R-tensors computable at once with the given cache size
        slices_z = np.floor(self.max_cache / (transition_matrix.nbytes / 1.e6)).astype(int)
        # Optimizer state
        state = AMMOptimizerState(expectations_selected, measurements, measurement_weights,
                                  stationary_distribution, slices_z, count_matrix_symmetric, counts_row_sums)
        ll_old = state.log_likelihood_biased(count_matrix, transition_matrix)

        state.log_likelihoods.append(ll_old)
        # make sure everything is initialized
        state.update_pi_hat()
        state.update_m_hat()
        state.update_Q()
        state.update_X_and_pi()

        ll_old = state.log_likelihood_biased(count_matrix, transition_matrix)
        state.log_likelihood_prev = ll_old
        state.update_G()

        #
        # Main estimation algorithm
        # 2-step algorithm, lagrange multipliers and pihat have different convergence criteria
        # when the lagrange multipliers have converged, pihat is updated until the log-likelihood has converged
        # (changes are smaller than 1e-3).
        # These do not always converge together, but usually within a few steps of each other.
        # A better heuristic for the latter may be necessary. For realistic cases (the two ubiquitin examples in [1])
        # this yielded results very similar to those with more stringent convergence criteria
        # (changes smaller than 1e-9) with convergence times
        # which are seconds instead of tens of minutes.
        #

        converged = False  # Convergence flag for lagrange multipliers
        i = 0
        die = False
        while i <= self.maxiter:
            pi_hat_old = state.pi_hat.copy()
            state.update_pi_hat()
            if not np.all(state.pi_hat > 0):
                state.pi_hat = pi_hat_old.copy()
                die = True
                self._log.warning("pihat does not have a finite probability for all states, terminating")
            state.update_m_hat()
            state.update_Q()

            if i > 1:
                X_old = np.copy(state.X)
                state.update_X_and_pi()
                if np.any(state.X[nonzero_counts] < 0) and i > 0:
                    die = True
                    self._log.warning(
                        "Warning: new X is not proportional to C... reverting to previous step and terminating")
                    state.X = X_old

            if not converged:
                self._newton_lagrange(state, count_matrix)
            else:  # once Lagrange multipliers are converged compute likelihood here
                transition_matrix = state.X / state.pi[:, None]
                _ll_new = state.log_likelihood_biased(count_matrix, transition_matrix)
                state.log_likelihoods.append(_ll_new)

            # General case fixed-point iteration
            if len(count_outside) > 0:
                if i > 1 and np.all((np.abs(state.delta_m_hat) / self.uncertainties) < self.convergence_criterion_lagrange)\
                        and not converged:
                    self._log.info(f"Converged Lagrange multipliers after {i} steps...")
                    converged = True
            # Special case
            else:
                if np.abs(state.log_likelihoods[-2] - state.log_likelihoods[-1]) < 1e-8:
                    self._log.info(f"Converged Lagrange multipliers after {i} steps...")
                    converged = True
            # if Lagrange multipliers are converged, check whether log-likelihood has converged
            if converged and np.abs(state.log_likelihoods[-2] - state.log_likelihoods[-1]) < 1e-8:
                self._log.info(f"Converged pihat after {i} steps...")
                die = True
            if die:
                break
            if i == self.maxiter:
                ll_diff = np.abs(state.log_likelihoods[-2] - state.log_likelihoods[-1])
                self._log.info(f"Failed to converge within {i} iterations. Log-likelihoods lastly changed by {ll_diff}."
                               f" Consider increasing max_iter(now={self.max_iter})")
            i += 1

        transition_matrix = msmest.transition_matrix(count_matrix, reversible=True, mu=state.pi_hat)
        self._model = AugmentedMSM(transition_matrix=transition_matrix, stationary_distribution=state.pi_hat,
                                   reversible=True, count_model=count_model, amm_optimizer_state=state)
        return self



from typing import Optional

import numpy as np

__author__ = "clonker"


class Ellipsoids:
    r""" Example data of a two-state markov chain which can be featurized into two parallel ellipsoids and optionally
    rotated into higher-dimensional space.

    Parameters
    ----------
    laziness : float in half-open interval (0.5, 1.], default=0.97
        The probability to stay in either state rather than transitioning. This yields a transition matrix of

        .. math:: P = \begin{pmatrix} \lambda & 1-\lambda \\ 1-\lambda & \lambda \end{pmatrix},

        where :math:`\lambda` is the selected laziness parameter.
    seed : int, optional, default=None
        Optional random seed for reproducibility.
    """

    state_0_mean = np.array([0., 0.])
    state_1_mean = np.array([0., 5.])

    def __init__(self, laziness: float = 0.97, seed: Optional[int] = None):
        if laziness <= 0.5 or laziness > 1:
            raise ValueError("Laziness must be at least 0.5 and at most 1.0 but was {}".format(laziness))
        transition_matrix = np.array([[laziness, 1-laziness], [1-laziness, laziness]])
        from deeptime.markov.msm import MarkovStateModel
        self._msm = MarkovStateModel(transition_matrix)
        self._rnd = np.random.RandomState(seed=seed)
        self._seed = seed
        self._cov = np.array([[5.7, 5.65], [5.65, 5.7]])

    @property
    def msm(self):
        r""" Yields the underlying markov state model. """
        return self._msm

    @property
    def random_state(self):
        r""" The random state that is used for RNG. """
        return self._rnd

    @property
    def seed(self):
        r""" Integer value of the seed or None. """
        return self._seed

    @property
    def covariance_matrix(self):
        r""" Covariance matrix that is used to parameterize a multivariate Gaussian distribution, resulting
        in the ellipsoidal shape. """
        return self._cov

    def discrete_trajectory(self, n_steps: int):
        r""" Generates a sequence of states 0 and 1 based on the internal markov state model.

        Parameters
        ----------
        n_steps : int
            The number of steps.

        Returns
        -------
        dtraj : (n_steps,) ndarray
            Time series of states.
        """
        if self.seed is not None:
            return self.msm.simulate(n_steps, seed=self.seed)
        else:
            return self.msm.simulate(n_steps)

    def observations(self, n_steps, n_dim=2, noise=False):
        r""" Generates an observation sequence in n_dim-dimensional space.

        Parameters
        ----------
        n_steps : int
            The number of observations.
        n_dim : int, default=2
            The dimension of the observation sequence. In case it is larger than 2, the 2-dimensional sequence
            is rotated with a random rotation matrix

            .. math::
                R = \begin{pmatrix} 0 & \cos(\alpha) & -\sin(\alpha) & \cdots \\
                                    0 & \sin(\alpha) & \cos(\alpha) & \ldots \end{pmatrix}
                \in\mathbb{R}^{2\times n}

            into :math:`n`-dimensional space, where :math:`\alpha\sim\mathcal{U}(0, 2\pi)`.
        noise : bool, default=False
            Optionally equips all observations with additional Gaussian noise of variance 0.2.

        Returns
        -------
        ftraj : (n_steps, n_dim) ndarray
            Observation trajectory.
        """
        if n_dim < 2:
            raise ValueError("Dimension must be at least 2 but was {}".format(n_dim))
        dtraj = self.discrete_trajectory(n_steps)
        return self.map_discrete_to_observations(dtraj, n_dim=n_dim, noise=noise)

    def map_discrete_to_observations(self, dtraj, n_dim=2, noise=False):
        r""" Maps a discrete trajectory (see :meth:`discrete_trajectory`) to an observation trajectory
        as described in :meth:`observations`)"""
        ftraj = np.empty((len(dtraj), 2))
        state_0_indices = np.where(dtraj == 0)[0]  # indices where the state is 0
        state_1_indices = np.where(dtraj == 1)[0]  # indices where the state is 1
        # fill allocated space with samples
        ftraj[state_0_indices] = self.random_state.multivariate_normal(
            mean=self.state_0_mean, cov=self.covariance_matrix, size=len(state_0_indices)
        )
        ftraj[state_1_indices] = self.random_state.multivariate_normal(
            mean=self.state_1_mean, cov=self.covariance_matrix, size=len(state_1_indices)
        )
        if n_dim == 2:
            # feature trajectory is already 2-dimensional
            if noise:
                ftraj += self.random_state.normal(scale=0.2, size=ftraj.shape)
            return ftraj
        else:
            # random rotation into higher-dimensional space
            angle = self.random_state.uniform(0, 2. * np.pi)
            rotation_matrix = np.zeros((2, n_dim))
            rotation_matrix[0, 1] = rotation_matrix[1, 2] = np.cos(angle)
            rotation_matrix[1, 1] = np.sin(angle)
            rotation_matrix[0, 2] = -np.sin(angle)

            ftraj = ftraj @ rotation_matrix
            assert ftraj.shape == (len(dtraj), n_dim)
            if noise:
                ftraj += self.random_state.normal(scale=0.2, size=ftraj.shape)
            return ftraj



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('tools', parent_package, top_path)

    config.add_subpackage('analysis')
    config.add_subpackage('analysis.dense')

    config.add_subpackage('estimation')

    config.add_subpackage('flux')

    config.add_extension('kahandot',
                         sources=['kahandot/kahandot_module.cpp'],
                         language='c++')

    return config



r"""This module provides dense implementations for the computation of
expectation values for a given transition matrix.

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import numpy as np
import scipy.sparse as sparse
from ._decomposition import rdl_decomposition
from ._stationary_vector import stationary_distribution


def expected_counts(p0, T, n):
    r"""Compute expected transition counts for Markov chain after n steps.

    Expected counts are computed according to ..math::

    E[C_{ij}^{(n)}]=\sum_{k=0}^{n-1} (p_0^T T^{k})_{i} p_{ij}

    For N<=M, the sum is computed via successive summation of the
    following matrix vector products, p_1^t=p_0^t
    T,...,p_n^t=P_{n-1}^t T. For n>M, the sum is computed using the
    eigenvalue decomposition of T and applying the expression for a
    finite geometric series to each of the eigenvalues.

    Parameters
    ----------
    p0 : (M,) ndarray
        Starting (probability) vector of the chain.
    T : (M, M) ndarray
        Transition matrix of the chain.
    n : int
        Number of steps to take from initial state.

    Returns
    --------
    EC : (M, M) ndarray
        Expected value for transition counts after n steps.

    """
    M = T.shape[0]
    if sparse.issparse(T) or n <= M:
        return ec_matrix_vector(p0, T, n)
    else:
        return ec_geometric_series(p0, T, n)


def expected_counts_stationary(T, n, mu=None):
    r"""Expected transition counts for Markov chain in equilibrium.

    Since mu is stationary for T we have

    .. math::

        E(C^{(N)})=N diag(mu)*T.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix.
    n : int
        Number of steps for chain.
    mu : (M,) ndarray (optional)
        Stationary distribution for T. If mu is not specified it will be
        computed via diagonalization of T.

    Returns
    -------
    EC : numpy array, shape=(n,n)
        Expected value for transition counts after a propagation of n steps.

    """
    if n <= 0:
        if sparse.issparse(T):
            EC = sparse.coo_matrix(T.shape, dtype=float)
        else:
            EC = np.zeros(T.shape)
        return EC
    else:
        if mu is None:
            mu = stationary_distribution(T, check_inputs=False)
        if sparse.issparse(T):
            D_mu = sparse.diags(mu, 0)
            EC = n * D_mu.dot(T)
        else:
            EC = n * mu[:, np.newaxis] * T
        return EC


def geometric_series(q, n):
    r"""
    Compute finite geometric series.

                                \frac{1-q^{n+1}}{1-q}   q \neq 1
        \sum_{k=0}^{n} q^{k}=
                                 n+1                     q  = 1

    Parameters
    ----------
    q : array-like
        The common ratio of the geometric series.
    n : int
        The number of terms in the finite series.

    Returns
    -------
    s : float or ndarray
        The value of the finite series.

    """
    q = np.asarray(q)
    if n < 0:
        raise ValueError('Finite geometric series is only defined for n>=0.')
    else:
        """q is scalar"""
        if q.ndim == 0:
            if q == 1:
                s = (n + 1) * 1.0
                return s
            else:
                s = (1.0 - q ** (n + 1)) / (1.0 - q)
                return s
        """q is ndarray"""
        s = np.zeros(np.shape(q), dtype=q.dtype)
        """All elements with value q=1"""
        ind = (q == 1.0)
        """For q=1 the sum has the value s=n+1"""
        s[ind] = (n + 1) * 1.0
        """All elements with value q\neq 1"""
        not_ind = np.logical_not(ind)
        s[not_ind] = (1.0 - q[not_ind] ** (n + 1)) / (1.0 - q[not_ind])
        return s


def ec_matrix_vector(p0, T, n):
    r"""Compute expected transition counts for Markov chain after n
    steps.

    Expected counts are computed according to ..math::

    E[C_{ij}^{(n)}]=\sum_{k=0}^{n-1} (p_0^t T^{k})_{i} p_{ij}

    The sum is computed via successive summation of the following
    matrix vector products, p_1^t=p_0^t T,...,p_n^t=P_{n-1}^t T.

    Such a direct approach can become prohibetively expensive for
    large n.  In this case the sum can be computed more efficiently
    using an eigenvalue decomposition of T and applying the closed
    form expression for a finite geometric series to each of the
    eigenvalues.

    Parameters
    ----------
    p0 : (M,) ndarray
        Starting (probability) vector of the chain.
    T : (M, M) ndarray
        Transition matrix of the chain.
    n : int
        Number of steps to take from initial state.

    Returns
    --------
    EC : (M, M) ndarray
        Expected value for transition counts after N steps.

    """
    if n <= 0:
        if sparse.issparse(T):
            return sparse.coo_matrix(T.shape, dtype=float)
        else:
            return np.zeros(T.shape)
    else:
        """Probability vector after (k=0) propagations"""
        p_k = 1.0 * p0
        """Sum of vectors after (k=0) propagations"""
        p_sum = 1.0 * p_k
        """Transpose T to use sparse dot product"""
        Tt = T.transpose()
        for k in range(n - 1):
            """Propagate one step p_{k} -> p_{k+1}"""
            p_k = Tt.dot(p_k)
            """Update sum"""
            p_sum += p_k
        """Expected counts"""
        if sparse.issparse(T):
            D_psum = sparse.diags(p_sum, 0)
            EC = D_psum.dot(T)
        else:
            EC = p_sum[:, np.newaxis] * T
        return EC


def ec_geometric_series(p0, T, n):
    r"""Compute expected transition counts for Markov chain after n
    steps.

    Expected counts are computed according to ..math::

    E[C_{ij}^{(n)}]=\sum_{k=0}^{n-1} (p_0^t T^{k})_{i} p_{ij}

    The sum is computed using the eigenvalue decomposition of T and
    applying the expression for a finite geometric series to each of
    the eigenvalues.

    For small n the computation of the eigenvalue decomposition can be
    much more expensive than a direct computation. In this case it is
    beneficial to compute the expected counts using successively
    computed matrix vector products p_1^t=p_0^t T, ... as increments.

    Parameters
    ----------
    p0 : (M,) ndarray
        Starting (probability) vector of the chain.
    T : (M, M) ndarray
        Transition matrix of the chain.
    n : int
        Number of steps to take from initial state.

    Returns
    --------
    EC : (M, M) ndarray
        Expected value for transition counts after N steps.

    """
    if n <= 0:
        EC = np.zeros(T.shape)
        return EC
    else:
        R, D, L = rdl_decomposition(T)
        w = np.diagonal(D)
        L = np.transpose(L)

        D_sum = np.diag(geometric_series(w, n - 1))
        T_sum = np.dot(np.dot(R, D_sum), np.conjugate(np.transpose(L)))
        p_sum = np.dot(p0, T_sum)
        EC = p_sum[:, np.newaxis] * T
        """Truncate imginary part - which is zero, but we want real
        return values"""
        EC = EC.real
        return EC



import numpy as np

from threadpoolctl import threadpool_limits

from . import _data_bindings as bd
from ..util.decorators import plotting_function
from ..util.parallel import handle_n_jobs, joining


class PBFSimulator:
    r""" A position based fluids simulator for two-dimensional systems. :footcite:`macklin2013position`

    Its underlying principle is by definition of a rest density :math:`\rho_0`, which the particles in the system
    try to reach by a smoothed particle hydrodynamics style simulation :footcite:`gingold1977smoothed`
    :footcite:`lucy1977numerical`. Up to numerics the simulation is deterministic given a set of parameters.

    Parameters
    ----------
    domain_size : (2,) ndarray
        A 1-dimensional ndarray with two elements describing the extension of the simulation box.
    initial_positions : (N, 2) ndarray
        A 2-dimensional ndarray describing the :math:`N` particles' initial positions. This also fixes the number
        of particles. It is preserved during runs of the simulation.
    interaction_distance : float
        Interaction distance for particles, influences the cell size for the neighbor list.
    n_jobs : int or None, default=None
        Number threads to use for simulation.
    n_solver_iterations : int, default=5
        The number of solver iterations for particle position updates. The more, the slower the simulation but also
        the more accurate.
    gravity : float, default=10
        Gravity parameter which acts on particles' velocities in each time step as constant force.
    epsilon : float, default=10
        Damping parameter. The higher, the more damping.
    timestep : float, default=0.016
        The timestep used for propagation.
    rest_density : float, defaul=1.
        The rest density :math:`\rho_0`.
    tensile_instability_distance : float, default=0.2
        Parameter responsible for surface-tension-like effects.
    tensile_instability_k : float, default=0.1
        Also controls surface-tension effects.

    Notes
    -----

    Each particle has positions :math:`p_1,\ldots,p_n \in\mathbb{R}^d` and
    velocity :math:`v_1,\ldots,v_n\in\mathbb{R^d}`. For the local density a standard SPH estimator is used

    .. math ::
        \rho_i = \sum_i m_i W(p_i - p_j, h),

    where :math:`m_i` refers to the particle's mass (here :math:`m_i \equiv 1` for simplicity), :math:`W` is a
    convolution kernel with interaction distance :math:`h`, i.e., particles with distance larger than :math:`h`
    get assigned a weight of zero in the convolution operation.

    The whole system has an underlying family of density constraints

    .. math ::
        C_i(p_1,\ldots,p_n) = \frac{\rho_i}{\rho_0} - 1.

    The simulation algorithm tries to fulfill this constraint by updating each particle's position, i.e., it tries to
    find :math:`\Delta p_k` so that

    .. math ::
        C_k(p_1,\ldots,p_k + \Delta p_k,\ldots,p_n) = 0.

    This is done by performing Newton's optimization method along the constraint's gradient. Because of potentially
    vanishing gradients, the optimization is regularized / dampened with a parameter :math:`\varepsilon`. In this
    implementation, larger values of :math:`\varepsilon` lead to more regularization.

    When a particle has only very few neighbors, it might not be possible to reach the rest density. This can lead
    to particle clumping (i.e., areas of negative pressures). One remedy is an artificial internal pressure term,
    here referred to as tensile instability. It modifies the weighting inside the convolution kernel itself.
    The default parameters `tensile_instability_distance` and `tensile_instability_k` are the suggested ones in the
    PBF publication.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, domain_size: np.ndarray, initial_positions: np.ndarray,
                 interaction_distance: float, n_jobs=None, n_solver_iterations: int = 5,
                 gravity: float = 10., epsilon: float = 10., timestep: float = 0.016, rest_density: float = 1.,
                 tensile_instability_distance: float = .2, tensile_instability_k: float = 0.1):
        n_jobs = handle_n_jobs(n_jobs)
        if np.atleast_1d(domain_size).ndim != 1 or np.atleast_1d(domain_size).shape[0] != 2 or np.any(domain_size <= 0):
            raise ValueError("Invalid domain size: must be positive and 1-dimensional of length two.")
        if initial_positions.ndim != 2 or initial_positions.shape[1] != 2:
            raise ValueError("initial positions must be a 2-dimensional numpy array of shape (N, 2), where N is"
                             "the number of particles.")
        if interaction_distance <= 0:
            raise ValueError("Interaction distance must be positive.")
        domain_size = domain_size.astype(np.float32, subok=False, copy=False)
        initial_positions = initial_positions.astype(np.float32, subok=False, copy=False)

        self._engine = bd.PBF(initial_positions, domain_size, interaction_distance, n_jobs)
        self._engine.n_solver_iterations = n_solver_iterations
        self._engine.gravity = gravity
        self._engine.epsilon = epsilon
        self._engine.timestep = timestep
        self._engine.rest_density = rest_density
        self._engine.tensile_instability_distance = tensile_instability_distance
        self._engine.tensile_instability_k = tensile_instability_k

    def run(self, n_steps: int, drift: float = 0.):
        r""" Performs `n_steps` many simulation steps and returns a trajectory of recorded particle positions.

        Parameters
        ----------
        n_steps : int
            Number of steps to run.
        drift : float, default=0
            Drift that is added to particles' velocity in `x` direction.

        Returns
        -------
        trajectory : (n_steps, 2) ndarray
            Output trajectory.
        """
        return self._engine.run(n_steps, drift)

    def simulate_oscillatory_force(self, n_oscillations, n_steps, drift=0.3):
        r""" Runs `2*n_oscillations*n_steps` many simulation steps in total by applying the drift alternatingly
        in positive and  negative direction for `n_steps` each, `n_oscillation` times.

        Parameters
        ----------
        n_oscillations : int
            Number of oscillations.
        n_steps : int
            Number of steps per run with a certain drift setting. Is then run again with the negated drift.
        drift : float
            The magnitude of the drift force into x direction.

        Returns
        -------
        trajectory : (T, n) ndarray
            Output trajectory.
        """
        n_runs = 2 * n_oscillations
        traj_total = np.empty((n_runs * n_steps, self.n_particles * 2))
        for i in range(n_runs):
            traj = self.run(n_steps, drift if i % 2 == 0 else -drift)
            traj_total[i * n_steps:(i + 1) * n_steps] = traj
        return traj_total

    def transform_to_density(self, trajectory, n_grid_x=20, n_grid_y=10, n_jobs=None):
        r"""Transforms a two-dimensional PBF particle trajectory to a trajectory of densities by performing KDEs.
        Since we have the prior knowledge that no particles get lost on the way, the densities are
        normalized frame-wise.

        Parameters
        ----------
        trajectory : (T, n, 2) ndarray
            The input trajectory for n particles.
        n_grid_x : int, default=20
            Number of evaluation points of simulation box in x direction.
        n_grid_y : int, default=10
            Number of evaluation points of simulation box in y direction.
        n_jobs : int or None, default=None
            Number of jobs to use when transforming to densities.

        Returns
        -------
        trajectory : (T, n_grid_x * n_grid_y) ndarray
            Output trajectory
        """
        n_jobs = handle_n_jobs(n_jobs)
        return _transform_to_density_impl(self.domain_size, trajectory, n_grid_x, n_grid_y, n_jobs)

    @plotting_function
    def make_animation(self, trajectories, stride=1, mode="scatter", **kw):  # pragma: no cover
        r""" Creates a matplotlib animation object consisting of either scatter plots of contour plots for plotting
        particles and densities, respectively.

        Parameters
        ----------
        trajectories : list of ndarray
            Input trajectories. Must all have same shape.
        stride : int, default=1
            Apply stride to frames so that fewer data is presented.
        mode : str, default="scatter"
            Aside from "scatter" this also supports "contourf" for densities.
        **kw
            Optional keyword arguments. The following are supported:

            * "n_grid_x", required when in mode "contourf" (see :meth:`transform_to_density`)
            * "n_grid_y", required when in mode "contourf" (see :meth:`transform_to_density`)
            * "figsize" controls figure size, defaults to (n_traj * 8, 6)
            * "ncols" controls number of columns, defaults to n_traj
            * "nrows" controls number of rows, defaults to 1
            * "titles" optional list of strings for titles

        Returns
        -------
        animation : FuncAnimation
            Matplotlib animation object.
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import matplotlib.animation as animation

        if not isinstance(trajectories, (list, tuple)):
            trajectories = [trajectories]
        trajectories = [traj.reshape((len(traj), -1, 2))[::stride] for traj in trajectories]

        n_particles = self.n_particles

        domain_size = self.domain_size

        agg_backend = kw.get('agg_backend', True)

        if agg_backend:
            backend_ = mpl.get_backend()
            mpl.use("Agg")  # Prevent showing stuff

        figsize = kw.get("figsize", (len(trajectories)*8, 6))
        ncols = kw.get("ncols", len(trajectories))
        nrows = kw.get("nrows", 1)
        titles = kw.get("titles", [None] * len(trajectories))
        fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
        if not isinstance(axes, np.ndarray):
            axes = np.asanyarray(axes)
        for title, ax in zip(titles, axes.flat):
            ax.set_xlim((-domain_size[0] / 2, domain_size[0] / 2))
            ax.set_ylim((-domain_size[1] / 2, domain_size[1] / 2))
            if title is not None:
                ax.set_title(title)

        s = np.empty((n_particles,))
        s.fill(300)
        c = np.empty((n_particles,))
        c.fill(0.5)
        plot_handles = []

        # needed for contourf
        grid = None
        gridx = None
        gridy = None

        if mode == "scatter":
            for traj, ax in zip(trajectories, axes.flat):
                handle = ax.scatter(traj[0, :, 0], traj[0, :, 1], s=s, c=c, vmin=0, vmax=1,
                                    cmap="jet", edgecolor="k")
                plot_handles.append(handle)
        elif mode == "contourf":
            n_grid_x = kw['n_grid_x']
            n_grid_y = kw['n_grid_y']

            gridx = np.linspace(-domain_size[0] / 2, domain_size[0] / 2, num=n_grid_x).astype(np.float32)
            gridy = np.linspace(-domain_size[1] / 2, domain_size[1] / 2, num=n_grid_y).astype(np.float32)
            grid = np.meshgrid(gridx, gridy)

            for k, ax in enumerate(axes.flat):
                handle = ax.contourf(grid[0], grid[1], trajectories[k][0].reshape((len(gridy), len(gridx))))
                plot_handles.append(handle)
        else:
            raise ValueError("Unknown mode {}".format(mode))

        def update_scatter(i):
            for traj, handle in zip(trajectories, plot_handles):
                X = traj[i]
                handle.set_offsets(X)
            return plot_handles

        def update_contourf(i):
            out = []
            for k, ax in enumerate(axes.flat):
                X = trajectories[k][i]
                for tp in plot_handles[k].collections:
                    tp.remove()
                plot_handles[k] = ax.contourf(grid[0], grid[1], X.reshape((len(gridy), len(gridx))))
                out += plot_handles[k].collections
            return out

        update = update_scatter if mode == "scatter" else update_contourf
        ani = animation.FuncAnimation(fig, update, interval=50, blit=True, repeat=False,
                                      frames=len(trajectories[0]))

        if agg_backend:
            mpl.use(backend_)  # Reset backend
        return ani

    @property
    def domain_size(self):
        r""" The size of the domain. """
        return self._engine.domain_size

    @property
    def n_particles(self):
        r""" Number of particles in the simulation. """
        return self._engine.n_particles


def _transform_to_density_impl_worker(args):
    with threadpool_limits(limits=1, user_api='blas'):
        t, frame, kde_input = args[0], args[1], args[2]
        from scipy.stats import gaussian_kde
        out = gaussian_kde(frame.T, bw_method=0.2).evaluate(kde_input.T)
        out /= out.sum()
        return t, out


def _transform_to_density_impl(domain_size, trajectory, n_grid_x=20, n_grid_y=10, n_jobs: int = 1):
    trajectory = trajectory.reshape((len(trajectory), -1, 2))

    gridx = np.linspace(-domain_size[0] / 2, domain_size[0] / 2, num=n_grid_x).astype(np.float32)
    gridy = np.linspace(-domain_size[1] / 2, domain_size[1] / 2, num=n_grid_y).astype(np.float32)
    grid = np.meshgrid(gridx, gridy)
    kde_input = np.dstack(grid).reshape(-1, 2)
    traj_kde = np.empty((len(trajectory), len(kde_input)))

    import multiprocessing as mp
    with joining(mp.Pool(processes=n_jobs)) as pool:
        args = [(t, trajectory[t], kde_input) for t in range(len(trajectory))]
        for result in pool.imap_unordered(_transform_to_density_impl_worker, args):
            traj_kde[result[0]] = result[1]

    return traj_kde



r"""This module provides functions for computation of prior count matrices

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

import numpy as np

from scipy.sparse import coo_matrix, issparse, csc_matrix


def prior_neighbor(C, alpha=0.001):
    r"""Neighbor prior of strength alpha for the given count matrix.

    Prior is defined by
        b_ij = alpha  if Z_ij+Z_ji > 0
        b_ij = 0      else

    Parameters
    ----------
    C : (M, M) scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) scipy.sparse matrix
        Prior count matrix

    """
    C_sym = C + C.transpose()
    C_sym = C_sym.tocoo()
    data = C_sym.data
    row = C_sym.row
    col = C_sym.col

    data_B = alpha * np.ones_like(data)
    B = coo_matrix((data_B, (row, col)))
    return B


def prior_const(C, alpha=0.001):
    """Constant prior of strength alpha.

    Prior is defined via

        b_ij=alpha for all i,j

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) ndarray
        Prior count matrix

    """
    B = alpha * np.ones(C.shape)
    return B


def prior_rev(C, alpha=-1.0):
    r"""Prior counts for sampling of reversible transition
    matrices.

    Prior is defined as

    b_ij= alpha if i<=j
    b_ij=0         else

    The reversible prior adds -1 to the upper triagular part of
    the given count matrix. This prior respects the fact that
    for a reversible transition matrix the degrees of freedom
    correspond essentially to the upper, respectively the lower
    triangular part of the matrix.

    Parameters
    ----------
    C : (M, M) ndarray or scipy.sparse matrix
        Count matrix
    alpha : float (optional)
        Value of prior counts

    Returns
    -------
    B : (M, M) ndarray
        Matrix of prior counts

    """
    ind = np.triu_indices(C.shape[0])
    if issparse(C):
        alphas = np.empty(len(ind[0]), dtype=np.float64)
        alphas.fill(alpha)
        B = csc_matrix((alphas, ind), shape=C.shape)
    else:
        B = np.zeros(C.shape, dtype=np.float64)
        B[ind] = alpha
    return B



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('numeric', parent_package, top_path)

    config.add_extension('eig_qr',
                         sources=['deeptime/numeric/eig_qr.pyx'],
                         )
    return config



import collections
from typing import List, Union, Optional

import numpy as np
from deeptime.markov._base import MembershipsChapmanKolmogorovValidator
from scipy.sparse import issparse

from ...base import Estimator
from .._transition_matrix import estimate_P, stationary_distribution
from ._hidden_markov_model import HiddenMarkovModel, viterbi
from ..msm import MarkovStateModel
from .. import TransitionCountModel, compute_dtrajs_effective
from ._hmm_bindings import util as _util
from ...util.types import ensure_timeseries_data


class MaximumLikelihoodHMM(Estimator):
    """ Maximum likelihood Hidden Markov model (HMM) estimator.

    This class is used to fit a maximum-likelihood HMM to data. It uses an initial guess HMM with can be obtained with
    one of the provided heuristics and then uses the Baum-Welch algorithm :footcite:`baum1967inequality` to fit
    the initial guess to provided data.

    The type of output model (gaussian or discrete)
    and the number of hidden states are extracted from the initial model. In case no initial distribution was given,
    the initial model assumes a uniform initial distribution.

    Parameters
    ----------
    initial_model : HiddenMarkovModel
        This model will be used to initialize the hidden markov model estimation routine. Since it is prone to
        get stuck in local optima, several initializations should be tried and scored and/or one of the available
        initialization heuristics should be applied, if appropriate.
    stride : int or str, optional, default=1
        stride between two lagged trajectories extracted from the input trajectories. Given trajectory s[t], stride
        and lag will result in trajectories

            :code:`s[0], s[lag], s[2 lag], ...`

            :code:`s[stride], s[stride + lag], s[stride + 2 lag], ...`

        Setting stride = 1 will result in using all data (useful for maximum likelihood estimator), while a
        Bayesian estimator requires a longer stride in order to have statistically uncorrelated trajectories.
        Setting stride = 'effective' uses the largest neglected timescale as an fit for the correlation time and
        sets the stride accordingly.
    lagtime : int, optional, default=1
        Lag parameter used for fitting the HMM
    reversible : bool, optional, default=True
        If True, a prior that enforces reversible transition matrices (detailed balance) is used;
        otherwise, a standard  non-reversible prior is used.
    stationary : bool, optional, default=False
        If True, the initial distribution of hidden states is self-consistently computed as the stationary
        distribution of the transition matrix. If False, it will be estimated from the starting states.
        Only set this to true if you're sure that the observation trajectories are initiated from a global
        equilibrium distribution.
    p : (n,) ndarray, optional, default=None
        Initial or fixed stationary distribution. If given and stationary=True, transition matrices will be
        estimated with the constraint that they have the set parameter as their stationary distribution.
        If given and stationary=False, the parameter is the fixed initial distribution of hidden states.
    accuracy : float, optional, default=1e-3
        Convergence threshold for EM iteration. When two the likelihood does not increase by more than
        accuracy, the iteration is stopped successfully.
    maxit : int, optional, default=1000
        Stopping criterion for EM iteration. When this many iterations are performed without reaching the requested
        accuracy, the iteration is stopped without convergence and a warning is given.
    maxit_reversible : int, optional, default=1000000
        Maximum number of iterations for reversible transition matrix estimation. Only used with reversible=True.

    References
    ----------
    .. footbibliography::
    """

    _HMMModelStorage = collections.namedtuple('_HMMModelStorage', ['transition_matrix', 'output_model',
                                                                   'initial_distribution'])

    def __init__(self, initial_model: HiddenMarkovModel, stride: Union[int, str] = 1,
                 lagtime: int = 1, reversible: bool = True, stationary: bool = False,
                 p: Optional[np.ndarray] = None, accuracy: float = 1e-3,
                 maxit: int = 1000, maxit_reversible: int = 100000):
        super().__init__()
        self.initial_model = initial_model
        self.stride = stride
        self.lagtime = lagtime
        self.reversible = reversible
        self.stationary = stationary
        if stationary:
            self.fixed_stationary_distribution = p
            self.fixed_initial_distribution = None
        else:
            self.fixed_stationary_distribution = None
            self.fixed_initial_distribution = p
        self.accuracy = accuracy
        self.maxit = maxit
        self.maxit_reversible = maxit_reversible

    def fetch_model(self) -> HiddenMarkovModel:
        r""" Yields the current HiddenMarkovModel or None if :meth:`fit` was not called yet.

        Returns
        -------
        model : HiddenMarkovModel or None
            The model.
        """
        return self._model

    @property
    def maxit_reversible(self) -> int:
        r"""Maximum number of iterations for reversible transition matrix estimation. Only used with reversible=True."""
        return self._maxit_reversible

    @maxit_reversible.setter
    def maxit_reversible(self, value: int):
        self._maxit_reversible = int(value)

    @property
    def fixed_stationary_distribution(self) -> Optional[np.ndarray]:
        r"""Fix the stationary distribution to the provided value. Only used when :attr:`stationary` is True, otherwise
        refer to :attr:`fixed_initial_distribution`.
        """
        return self._fixed_stationary_distribution

    @fixed_stationary_distribution.setter
    def fixed_stationary_distribution(self, value: Optional[np.ndarray]):
        if value is not None and value.shape[0] != self.n_hidden_states:
            raise ValueError("Fixed stationary distribution must be as long as there are hidden states.")
        self._fixed_stationary_distribution = value

    @property
    def fixed_initial_distribution(self) -> Optional[np.ndarray]:
        r"""Fix the initial distribution to the provided value. Only used when :attr:`stationary` is False, otherwise
        refer to :attr:`fixed_stationary_distribution`.
        """
        return self._fixed_initial_distribution

    @fixed_initial_distribution.setter
    def fixed_initial_distribution(self, value: Optional[np.ndarray]):
        if value is not None and value.shape[0] != self.n_hidden_states:
            raise ValueError("Fixed initial distribution must be as long as there are hidden states.")
        self._fixed_initial_distribution = value

    @property
    def stationary(self) -> bool:
        r""" If True, the initial distribution of hidden states is self-consistently computed as the stationary
        distribution of the transition matrix. If False, it will be estimated from the starting states.
        Only set this to true if you're sure that the observation trajectories are initiated from a global
        equilibrium distribution.
        """
        return self._stationary

    @stationary.setter
    def stationary(self, value: bool):
        self._stationary = bool(value)

    @property
    def accuracy(self) -> float:
        r""" Convergence threshold for EM iteration. """
        return self._accuracy

    @accuracy.setter
    def accuracy(self, value: float):
        self._accuracy = float(value)

    @property
    def maxit(self) -> int:
        r""" Stopping criterion for EM iteration. """
        return self._maxit

    @maxit.setter
    def maxit(self, value: int):
        self._maxit = int(value)

    @property
    def reversible(self) -> bool:
        r""" Whether the hidden transition model should be estimated so that it is reversible. """
        return self._reversible

    @reversible.setter
    def reversible(self, value: bool):
        self._reversible = bool(value)

    @property
    def stride(self) -> Union[int, str]:
        r""" Stride to be applied to the input data. Must be compatible with how the initial model was estimated. """
        return self._stride

    @stride.setter
    def stride(self, value):
        if isinstance(value, str):
            if value != 'effective':
                raise ValueError("stride value can only be either integer or 'effective'.")
            else:
                self._stride = value
        else:
            self._stride = int(value)

    @property
    def lagtime(self) -> int:
        r""" The lag time at which transitions are counted. """
        return self._lagtime

    @lagtime.setter
    def lagtime(self, value: int):
        value = int(value)
        if value <= 0:
            raise ValueError("Lagtime must be positive!")
        self._lagtime = value

    @property
    def n_hidden_states(self) -> int:
        r""" The number of hidden states, coincides with the number of hidden states in the initial model."""
        return self.initial_model.n_hidden_states

    @property
    def initial_model(self) -> HiddenMarkovModel:
        r""" The initial transition model. """
        return self._initial_transition_model

    @initial_model.setter
    def initial_model(self, value: HiddenMarkovModel) -> None:
        self._initial_transition_model = value

    def fit(self, dtrajs, initial_model=None, **kwargs):
        r""" Fits a new :class:`HMM <HiddenMarkovModel>` to data.

        Parameters
        ----------
        dtrajs : array_like or list of array_like
            Timeseries data.
        initial_model : HiddenMarkovModel, optional, default=None
            Override for :attr:`initial_model`.
        **kwargs
            Ignored kwargs for scikit-learn compatibility.

        Returns
        -------
        self : MaximumLikelihoodHMM
            Reference to self.
        """
        if initial_model is None:
            initial_model = self.initial_model
        if initial_model is None or not isinstance(initial_model, HiddenMarkovModel):
            raise ValueError("For estimation, an initial model of type "
                             "`deeptime.markov.hmm.HiddenMarkovModel` is required.")

        # copy initial model
        transition_matrix = initial_model.transition_model.transition_matrix
        if issparse(transition_matrix):
            # want dense matrix, toarray makes a copy
            transition_matrix = transition_matrix.toarray()
        else:
            # new instance
            transition_matrix = np.copy(transition_matrix)

        hmm_data = MaximumLikelihoodHMM._HMMModelStorage(transition_matrix=transition_matrix,
                                                         output_model=initial_model.output_model.copy(),
                                                         initial_distribution=initial_model.initial_distribution.copy())

        dtrajs = ensure_timeseries_data(dtrajs)
        dtrajs = compute_dtrajs_effective(dtrajs, lagtime=self.lagtime, n_states=initial_model.n_hidden_states,
                                          stride=self.stride)

        max_n_frames = max(len(obs) for obs in dtrajs)
        # pre-construct hidden variables
        N = initial_model.n_hidden_states
        alpha = np.zeros((max_n_frames, N))
        beta = np.zeros((max_n_frames, N))
        gammas = [np.zeros((len(obs), N)) for obs in dtrajs]
        count_matrices = [np.zeros((N, N)) for _ in dtrajs]

        it = 0
        likelihoods = np.empty(self.maxit)
        # flag if connectivity has changed (e.g. state lost) - in that case the likelihood
        # is discontinuous and can't be used as a convergence criterion in that iteration.
        tmatrix_nonzeros = hmm_data.transition_matrix.nonzero()
        converged = False

        while not converged and it < self.maxit:
            loglik = 0.0
            for obs, gamma, counts in zip(dtrajs, gammas, count_matrices):
                loglik_update, _ = self._forward_backward(hmm_data, obs, alpha, beta, gamma, counts)
                loglik += loglik_update
            assert np.isfinite(loglik), it

            # convergence check
            if it > 0:
                dL = loglik - likelihoods[it - 1]
                if dL < self.accuracy:
                    converged = True

            # update model
            self._update_model(hmm_data, dtrajs, gammas, count_matrices, maxiter=self.maxit_reversible)

            # connectivity change check
            tmatrix_nonzeros_new = hmm_data.transition_matrix.nonzero()
            if not np.array_equal(tmatrix_nonzeros, tmatrix_nonzeros_new):
                converged = False  # unset converged
                tmatrix_nonzeros = tmatrix_nonzeros_new

            # end of iteration
            likelihoods[it] = loglik
            it += 1

        likelihoods = np.resize(likelihoods, it)

        transition_counts = self._reduce_transition_counts(count_matrices)

        count_model = TransitionCountModel(count_matrix=transition_counts, lagtime=self.lagtime)
        transition_model = MarkovStateModel(hmm_data.transition_matrix, reversible=self.reversible,
                                            count_model=count_model)
        hidden_state_trajs = [
            viterbi(hmm_data.transition_matrix, hmm_data.output_model.to_state_probability_trajectory(obs),
                    hmm_data.initial_distribution) for obs in dtrajs
        ]
        model = HiddenMarkovModel(
            transition_model=transition_model,
            output_model=hmm_data.output_model,
            initial_distribution=hmm_data.initial_distribution,
            likelihoods=likelihoods,
            state_probabilities=gammas,
            initial_count=self._init_counts(gammas),
            hidden_state_trajectories=hidden_state_trajs,
            stride=self.stride
        )
        self._model = model
        return self

    @staticmethod
    def _forward_backward(model: _HMMModelStorage, obs, alpha, beta, gamma, counts):
        """ Estimation step: Runs the forward-back algorithm on trajectory obs

        Parameters
        ----------
        model: _HMMModelStorage
            named tuple with transition matrix, initial distribution, output model
        obs: np.ndarray
            single observation corresponding to index itraj
        alpha: ndarray
            forward coefficients
        beta: ndarray
            backward coefficients
        gamma: ndarray
            gammas
        counts: ndarray
            count matrix

        Returns
        -------
        logprob : float
            The probability to observe the observation sequence given the HMM
            parameters
        pobs : ndarray
            state probability trajectory obtained from obs
        """
        # get parameters
        A = model.transition_matrix
        pi = model.initial_distribution
        T = len(obs)
        # compute output probability matrix
        pobs = model.output_model.to_state_probability_trajectory(obs)
        pobs = pobs.astype(A.dtype)
        # run forward - backward pass
        logprob = _util.forward_backward(A, pobs, pi, alpha, beta, gamma, counts, T)
        return logprob, pobs

    def _init_counts(self, gammas):
        gamma0_sum = np.zeros(self.n_hidden_states)
        # update state counts
        for g in gammas:
            gamma0_sum += g[0]
        return gamma0_sum

    @staticmethod
    def _reduce_transition_counts(count_matrices):
        return np.add.reduce(count_matrices)

    def _update_model(self, model: _HMMModelStorage, observations: List[np.ndarray], gammas: List[np.ndarray],
                      count_matrices: List[np.ndarray], maxiter: int = int(1e7)):
        """
        Maximization step: Updates the HMM model given the hidden state assignment and count matrices

        Parameters
        ----------
        gammas : [ ndarray(T,N, dtype=float) ]
            list of state probabilities for each trajectory
        count_matrices : [ ndarray(N,N, dtype=float) ]
            list of the Baum-Welch transition count matrices for each hidden
            state trajectory
        maxiter : int
            maximum number of iterations of the transition matrix estimation if
            an iterative method is used.

        """
        C = self._reduce_transition_counts(count_matrices)

        # compute new transition matrix
        T = estimate_P(C, reversible=self.reversible, fixed_statdist=self.fixed_stationary_distribution,
                       maxiter=maxiter, maxerr=1e-12, mincount_connectivity=1e-16)
        # estimate stationary or init distribution
        if self.stationary:
            if self.fixed_stationary_distribution is None:
                pi = stationary_distribution(T, C=C, mincount_connectivity=1e-16)
            else:
                pi = self.fixed_stationary_distribution
        else:
            if self.fixed_initial_distribution is None:
                gamma0_sum = self._init_counts(gammas)
                pi = gamma0_sum / np.sum(gamma0_sum)
            else:
                pi = self.fixed_initial_distribution

        model.initial_distribution[:] = pi
        model.transition_matrix[:] = T
        model.output_model.fit(observations, gammas)

    def chapman_kolmogorov_validator(self, mlags, test_model: HiddenMarkovModel = None):
        r""" Creates a validator instance which can be used to perform a Chapman-Kolmogorov test.

        Parameters
        ----------
        mlags : int or int-array
            Multiples of lag times for testing the Model, e.g. range(10).
            A single int will trigger a range, i.e. mlags=10 maps to
            mlags=range(1, 10). The setting None will choose mlags automatically
            according to the longest available trajectory.
        test_model : HiddenMarkovModel, optional, default=None
            The model that is tested. If not provided, uses this estimator's encapsulated model.

        Returns
        -------
        validator : MembershipsChapmanKolmogorovValidator
            The validator.

        Raises
        ------
        AssertionError
            If test_model is None and this estimator has not been :meth:`fit` on data yet or the output model
            was not a discrete output model.
        """
        test_model = self.fetch_model() if test_model is None else test_model
        assert test_model is not None, "We need a test model via argument or an estimator which was already" \
                                       "fit to data."
        from . import DiscreteOutputModel
        assert isinstance(test_model.output_model, DiscreteOutputModel), \
            "Can only perform CKTest for discrete output models"
        lagtime = test_model.lagtime
        return MLHMMChapmanKolmogorovValidator(test_model, self, np.eye(test_model.n_hidden_states), lagtime, mlags)


def _ck_estimate_model_for_lag(estimator: MaximumLikelihoodHMM, model: HiddenMarkovModel, data, lagtime):
    from .init.discrete import metastable_from_data
    initial_model = metastable_from_data(data, n_hidden_states=model.n_hidden_states, lagtime=lagtime,
                                         stride=estimator.stride, reversible=estimator.reversible,
                                         stationary=estimator.stationary)
    estimator = MaximumLikelihoodHMM(initial_model, lagtime=lagtime, reversible=estimator.reversible,
                                     stationary=estimator.stationary, accuracy=estimator.accuracy,
                                     maxit=estimator.maxit, maxit_reversible=estimator.maxit_reversible)
    hmm = estimator.fit(data).fetch_model()
    return hmm.submodel_largest(dtrajs=data)


class MLHMMChapmanKolmogorovValidator(MembershipsChapmanKolmogorovValidator):

    def fit(self, data, n_jobs=1, progress=None, estimate_model_for_lag=None, **kw):
        if n_jobs != 1:
            import warnings
            warnings.warn("Ignoring n_jobs for hmm cktest")
        return super().fit(data, n_jobs=1, estimate_model_for_lag=_ck_estimate_model_for_lag, progress=progress, **kw)



import warnings

import numpy as np

from . import metrics
from ._clustering_bindings import regspace as _regspace_ext
from ._cluster_model import ClusterModel
from ..base import Estimator

__all__ = ['RegularSpace']

from ..util.parallel import handle_n_jobs


class RegularSpace(Estimator):
    """Clusters data objects in such a way, that cluster centers are at least in distance of dmin to each other
    according to the given metric. The assignment of data objects to cluster centers is performed by Voronoi partioning.

    Regular space clustering :footcite:`prinz2011markov` is very similar to Hartigan's leader
    algorithm :footcite:`hartigan1975clustering`. It consists
    of two passes through the data. Initially, the first data point is added to the list of centers. For every
    subsequent data point, if it has a greater distance than dmin from every center, it also becomes a center.
    In the second pass, a Voronoi discretization with the computed centers is used to partition the data.

    Parameters
    ----------
    dmin : float
        Minimum distance between all clusters, must be non-negative.
    max_centers : int
        If this threshold is met during finding the centers, the algorithm will terminate. Must be positive.
    metric : str, default='euclidean'
        The metric to use during clustering. For a list of available metrics,
        see the :data:`metric registry <deeptime.clustering.metrics>`.
    n_jobs : int, optional, default=None
        Number of threads to use during estimation.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, dmin: float, max_centers: int = 1000, metric: str = 'euclidean', n_jobs=None):
        super(RegularSpace, self).__init__()
        self.dmin = dmin
        self.metric = metric
        self.max_centers = max_centers
        self.n_jobs = n_jobs
        self._clustercenters = []
        self._converged = False

    @property
    def metric(self) -> str:
        r"""
        The metric that is used for clustering.

        :type: str.
        """
        return self._metric

    @metric.setter
    def metric(self, value: str):
        if value not in metrics.available:
            raise ValueError(f"Unknown metric {value}, available metrics: {metrics.available}")
        self._metric = value

    @property
    def dmin(self) -> float:
        r"""
        Minimum distance between cluster centers.

        :getter: Yields the currently set minimum distance.
        :setter: Sets a new minimum distance, must be non-negative.
        :type: float
        """
        return self._dmin

    @dmin.setter
    def dmin(self, d: float):
        if d < 0:
            raise ValueError("d has to be non-negative")

        self._dmin = d

    @property
    def max_centers(self) -> int:
        r"""
        Cutoff during clustering. If reached no more data is taken into account. You might then consider a larger
        value or a larger dmin value.

        :getter: Current maximum number of cluster centers.
        :setter: Sets a new maximum number of cluster centers, must be non-negative.
        :type: int
        """
        return self._max_centers

    @max_centers.setter
    def max_centers(self, value: int):
        if value <= 0:
            raise ValueError("max_centers has to be positive")

        self._max_centers = value

    @property
    def n_clusters(self) -> int:
        r""" Alias to :attr:`max_centers`. """
        return self.max_centers

    @n_clusters.setter
    def n_clusters(self, val: int):
        self.max_centers = val

    @property
    def n_jobs(self) -> int:
        r"""
        The number of threads to use during estimation.

        :getter: Yields the number of threads to use, -1 is an allowed value for all available threads.
        :setter: Sets the number of threads to use, can be None in which case it defaults to 1 thread.
        :type: int
        """
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, value):
        self._n_jobs = handle_n_jobs(value)

    def fetch_model(self) -> ClusterModel:
        """
        Fetches the current model. Can be `None` in case :meth:`fit` was not called yet.

        Returns
        -------
        model : ClusterModel or None
            The latest estimated model or None.
        """
        clustercenters = np.array([])
        if len(self._clustercenters) > 0:
            dim = len(self._clustercenters[0][0])
            clustercenters = np.asarray_chkfinite(self._clustercenters).reshape(-1, dim)
        self._model = ClusterModel(clustercenters, self.metric, self._converged)
        return self._model

    def partial_fit(self, data, n_jobs=None):
        r""" Fits data to an existing model. See :meth:`fit`. """
        n_jobs = self.n_jobs if n_jobs is None else handle_n_jobs(n_jobs)
        if data.ndim == 1:
            data = data[:, np.newaxis]
        try:
            metric = metrics[self.metric]()
            _regspace_ext.cluster(data, self._clustercenters, self.dmin, self.max_centers, n_jobs, metric)
            self._converged = True
        except _regspace_ext.MaxCentersReachedException:
            warnings.warn('Maximum number of cluster centers reached.'
                          ' Consider increasing max_centers or choose'
                          ' a larger minimum distance, dmin.')

        return self

    def fit(self, data, n_jobs=None):
        r"""
        Fits this estimator onto data. The estimation is carried out by

        #. Choosing first data frame as centroid
        #. for all frames :math:`x\in X`: Calculate distance to all cluster centers
        #. Add a new centroid if minimal distance to all other cluster centers is larger or equal :attr:`dmin`.

        Parameters
        ----------
        data : (T, n) ndarray or list of ndarray
            the data to fit
        n_jobs : int, optional, default=None
            Number of jobs, superseeds :attr:`n_jobs` if set to an integer value

        Returns
        -------
        self : RegularSpace
            reference to self
        """
        self._clustercenters.clear()
        self._converged = False

        if not isinstance(data, (tuple, list)):
            data = [data]

        for traj in data:
            self.partial_fit(traj, n_jobs=n_jobs)

        return self



import numpy as np

from ._util import compute_connected_sets, closed_sets, is_connected


def estimate_P(C, reversible=True, fixed_statdist=None, maxiter=1000000, maxerr=1e-8, mincount_connectivity=0):
    """ Estimates full transition matrix for general connectivity structure

    Parameters
    ----------
    C : ndarray
        count matrix
    reversible : bool
        estimate reversible?
    fixed_statdist : ndarray or None
        estimate with given stationary distribution
    maxiter : int
        Maximum number of reversible iterations.
    maxerr : float
        Stopping criterion for reversible iteration: Will stop when infinity
        norm  of difference vector of two subsequent equilibrium distributions
        is below maxerr.
    mincount_connectivity : float
        Minimum count which counts as a connection.

    """
    import deeptime.markov.tools.estimation as msmest
    n = np.shape(C)[0]
    # output matrix. Set initially to Identity matrix in order to handle empty states
    P = np.eye(n, dtype=np.float64)
    # decide if we need to proceed by weakly or strongly connected sets
    if reversible and fixed_statdist is None:  # reversible to unknown eq. dist. - use strongly connected sets.
        S = compute_connected_sets(C, connectivity_threshold=mincount_connectivity, directed=True)
        for s in S:
            mask = np.zeros(n, dtype=bool)
            mask[s] = True
            if C[np.ix_(mask, ~mask)].sum() > np.finfo(C.dtype).eps:  # outgoing transitions - use partial rev algo.
                transition_matrix_partial_rev(C, P, mask, maxiter=maxiter, maxerr=maxerr)
            else:  # closed set - use standard estimator
                indices = np.ix_(mask, mask)
                if s.size > 1:  # leave diagonal 1 if single closed state.
                    P[indices] = msmest.transition_matrix(C[indices], reversible=True, warn_not_converged=False,
                                                    maxiter=maxiter, maxerr=maxerr)
    else:  # nonreversible or given equilibrium distribution - weakly connected sets
        S = compute_connected_sets(C, connectivity_threshold=mincount_connectivity, directed=False)
        for s in S:
            indices = np.ix_(s, s)
            if not reversible:
                Csub = C[indices]
                # any zero rows? must set Cii = 1 to avoid dividing by zero
                zero_rows = np.where(Csub.sum(axis=1) == 0)[0]
                Csub[zero_rows, zero_rows] = 1.0
                P[indices] = msmest.transition_matrix(Csub, reversible=False)
            elif reversible and fixed_statdist is not None:
                P[indices] = msmest.transition_matrix(C[indices], reversible=True, fixed_statdist=fixed_statdist,
                                                maxiter=maxiter, maxerr=maxerr)
            else:  # unknown case
                raise NotImplementedError('Transition estimation for the case reversible=' + str(reversible) +
                                          ' fixed_statdist=' + str(fixed_statdist is not None) + ' not implemented.')
    # done
    return P


def transition_matrix_partial_rev(C, P, S, maxiter=1000000, maxerr=1e-8):
    r"""Maximum likelihood estimation of transition matrix which is reversible on parts

    Partially-reversible estimation of transition matrix. Maximizes the likelihood:

    .. math:
        P_S &=& arg max prod_{S, :} (p_ij)^c_ij \\
        \\Pi_S P_{S,S} &=& \\Pi_S P_{S,S}

    where the product runs over all elements of the rows S, and detailed balance only
    acts on the block with rows and columns S. :math:`\Pi_S` is the diagonal matrix of
    equilibrium probabilities restricted to set S.

    Note that this formulation

    Parameters
    ----------
    C : ndarray
        full count matrix
    P : ndarray
        full transition matrix to write to. Will overwrite P[S]
    S : ndarray, bool
        boolean selection of reversible set with outgoing transitions
    maxiter : int, optional, default = 1000000
        Maximum number of iterations, iteration termination condition.
    maxerr : float
        maximum difference in matrix sums between iterations (infinity norm) in order to stop.
    """
    # test input
    assert np.array_equal(C.shape, P.shape)
    # constants
    A = C[S][:, S]
    B = C[S][:, ~S]
    ATA = A + A.T
    countsums = C[S].sum(axis=1)
    # initialize
    X = 0.5 * ATA
    Y = C[S][:, ~S]
    # normalize X, Y
    totalsum = X.sum() + Y.sum()
    X /= totalsum
    Y /= totalsum
    # rowsums
    rowsums = X.sum(axis=1) + Y.sum(axis=1)
    err = 1.0
    it = 0
    while err > maxerr and it < maxiter:
        # update
        d = countsums / rowsums
        X = ATA / (d[:, None] + d)
        Y = B / d[:, None]
        # normalize X, Y
        totalsum = X.sum() + Y.sum()
        X /= totalsum
        Y /= totalsum
        # update sums
        rowsums_new = X.sum(axis=1) + Y.sum(axis=1)
        # compute error
        err = np.max(np.abs(rowsums_new - rowsums))
        # update
        rowsums = rowsums_new
        it += 1
    # write to P
    P[np.ix_(S, S)] = X
    P[np.ix_(S, ~S)] = Y
    P[S] /= P[S].sum(axis=1)[:, None]


def enforce_reversible_on_closed(P):
    """ Enforces transition matrix P to be reversible on its closed sets. """
    import deeptime.markov.tools.analysis as msmana
    Prev = P.copy()
    # treat each weakly connected set separately
    sets = closed_sets(P)
    for s in sets:
        indices = np.ix_(s, s)
        # compute stationary probability
        pi_s = msmana.stationary_distribution(P[indices])
        # symmetrize
        X_s = pi_s[:, None] * P[indices]
        X_s = 0.5 * (X_s + X_s.T)
        # normalize
        Prev[indices] = X_s / X_s.sum(axis=1)[:, None]
    return Prev


def is_reversible(P):
    """ Returns if P is reversible on its weakly connected sets """
    import deeptime.markov.tools.analysis as msmana
    # treat each weakly connected set separately
    sets = compute_connected_sets(P, directed=False)
    for s in sets:
        Ps = P[s, :][:, s]
        if not msmana.is_transition_matrix(Ps):
            return False  # isn't even a transition matrix!
        pi = msmana.stationary_distribution(Ps)
        X = pi[:, None] * Ps
        if not np.allclose(X, X.T):
            return False
    # survived.
    return True


def stationary_distribution(P, C=None, mincount_connectivity=0):
    """ Simple estimator for stationary distribution for multiple strongly connected sets """
    # can be replaced by deeptime.markov.tools.analysis.stationary_distribution in next msmtools release
    from deeptime.markov.tools.analysis import stationary_distribution as msmstatdist
    if C is None:
        if is_connected(P, directed=True):
            return msmstatdist(P, check_inputs=False)
        else:
            raise ValueError('Computing stationary distribution for disconnected matrix. Need count matrix.')

    # disconnected sets
    n = np.shape(C)[0]
    ctot = np.sum(C)
    pi = np.zeros(n)
    # treat each weakly connected set separately
    sets = compute_connected_sets(C, connectivity_threshold=mincount_connectivity, directed=False)
    for s in sets:
        # compute weight
        w = np.sum(C[s, :]) / ctot
        pi[s] = w * msmstatdist(P[s, :][:, s], check_inputs=False)
    # reinforce normalization
    pi /= np.sum(pi)
    return pi
